import os
import re
from pathlib import Path
from unittest.mock import MagicMock

import pytest

from briefcase.console import LogLevel
from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations import windows_sdk
from briefcase.integrations.windows_sdk import WindowsSDK


@pytest.fixture
def mock_winreg(monkeypatch):
    """Mock out winreg module."""
    winreg = MagicMock()
    monkeypatch.setattr(windows_sdk, "winreg", winreg)

    # Default all registry reads to "key doesn't exist"
    winreg.OpenKeyEx.side_effect = MagicMock()
    winreg.QueryValueEx.side_effect = FileNotFoundError

    # Initialize necessary but inconsequential winreg constants for testing
    winreg.KEY_READ = 131097
    winreg.KEY_WOW64_32KEY = 512
    winreg.KEY_WOW64_64KEY = 256
    winreg.HKEY_LOCAL_MACHINE = "HKEY_LOCAL_MACHINE"
    winreg.HKEY_CURRENT_USER = "HKEY_CURRENT_USER"
    winreg.REG_SZ = 1

    return winreg


def setup_winsdk_install(
    base_path: Path,
    version: str,
    arch="x64",
    skip_bins=False,
) -> (Path, str):
    """Create a mock Windows SDK for the version and arch.

    :param base_path: base path to create the SDK in; should be pytest's tmp_path.
    :param version: SDK version triple, e.g. 1.2.3. The created directory path will include
        a servicing version of 0, e.g. base_path/win_sdk/1.2.3.0.
    :param arch: The architecture for the SDK, e.g. amd64 or arm64.
    :param skip_bins: Do not create mock binaries in `bin` directory.
    :returns: tuple of path to base of SDK install and version triple
    """
    sdk_path = base_path / "win_sdk"
    sdk_ver = version
    (sdk_path / "bin" / f"{sdk_ver}.0" / arch).mkdir(parents=True, exist_ok=True)
    # Mock the necessary tools in the SDK
    if not skip_bins:
        (sdk_path / "bin" / f"{sdk_ver}.0" / arch / "signtool.exe").touch()
    return sdk_path, sdk_ver


def test_short_circuit(mock_tools):
    """Tool is not created if already cached."""
    mock_tools.windows_sdk = "tool"

    tool = WindowsSDK.verify(mock_tools)

    assert tool == "tool"
    assert tool == mock_tools.windows_sdk


@pytest.mark.parametrize("host_os", ["Darwin", "Linux", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{WindowsSDK.name} is not supported on {host_os}",
    ):
        WindowsSDK.verify(mock_tools)


@pytest.mark.parametrize(
    "host_arch, sdk_arch", [("AMD64", "x64"), ("ARM64", "arm64"), ("gothic", "gothic")]
)
def test_winsdk_arch(
    mock_tools,
    mock_winreg,
    host_arch,
    sdk_arch,
    tmp_path,
    monkeypatch,
):
    """The architecture of the host machine is respected."""
    # Mock the environment for a Windows SDK install
    sdk_path, sdk_ver = setup_winsdk_install(tmp_path, "1.1.1", sdk_arch)

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "1.1")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Mock lookup of Windows SDK
    mock_sdks = MagicMock(spec=WindowsSDK._windows_sdks)
    mock_sdks.return_value = [(sdk_path, f"{sdk_ver}.0")]
    monkeypatch.setattr(WindowsSDK, "_windows_sdks", mock_sdks)

    # Mock the machine
    mock_tools.host_arch = host_arch

    # Verify the install
    win_sdk = WindowsSDK.verify(mock_tools)

    # The returned paths are as expected (and are the full paths)
    signtool_path = tmp_path / "win_sdk/bin/1.1.1.0" / sdk_arch / "signtool.exe"
    assert win_sdk.signtool_exe == signtool_path


def test_winsdk_valid_env_vars(mock_tools, mock_winreg, tmp_path, monkeypatch):
    """If the WindowsSDKDir and WindowsSDKVersion env vars point to a suitable Windows
    SDK install, the validator succeeds."""
    # Mock the environment for a Windows SDK install
    sdk_path, sdk_ver = setup_winsdk_install(tmp_path, "1.1.1", "x64")

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "1.1")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Set up environment variables for SDK location
    mock_tools.os.environ = {
        "WindowsSDKDir": os.fsdecode(sdk_path),
        "WindowsSDKVersion": f"{sdk_ver}.0",
    }

    # Verify the install
    win_sdk = WindowsSDK.verify(mock_tools)

    # The returned paths are as expected (and are the full paths)
    signtool_path = tmp_path / "win_sdk/bin/1.1.1.0/x64/signtool.exe"
    assert win_sdk.signtool_exe == signtool_path


def test_winsdk_invalid_env_vars(mock_tools, mock_winreg, tmp_path, monkeypatch):
    """If the WindowsSDKDir and WindowsSDKVersion env vars point to an invalid Windows
    SDK install, the validator fails."""
    # Mock the environment for a Windows SDK install
    sdk_path = tmp_path / "win_sdk"
    sdk_ver = "1.1.1.0"

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "1.1")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Set up environment variables for SDK location
    mock_tools.os.environ = {
        "WindowsSDKDir": os.fsdecode(sdk_path),
        "WindowsSDKVersion": f"{sdk_ver}.0",
    }

    # Fail validation for missing install from env vars
    with pytest.raises(
        BriefcaseCommandError,
        match="The 'WindowsSDKDir' and 'WindowsSDKVersion' environment variables do not point",
    ):
        WindowsSDK.verify(mock_tools)


def test_winsdk_latest_install_from_reg(mock_tools, mock_winreg, tmp_path, monkeypatch):
    """If the first (i.e. "latest") version of the SDK found in the registry is valid,
    the validator succeeds."""
    # Bypass using SDK in env vars
    mock_tools.os.environ.get.return_value = None

    # Turn on verbose logging
    mock_tools.logger.verbosity = LogLevel.DEBUG

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "83.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Mock the SDK install
    sdk_path, sdk_ver = setup_winsdk_install(tmp_path, "83.0.1")

    # Return "latest" SDK as first match
    mock_winreg.QueryValueEx.side_effect = [
        (str(sdk_path), mock_winreg.REG_SZ),
        (sdk_ver, mock_winreg.REG_SZ),
    ]

    # Verify the install
    win_sdk = WindowsSDK.verify(mock_tools)

    # The environment was queried.
    mock_tools.os.environ.get.assert_called_once_with("WindowsSDKDir")

    # The returned paths are as expected (and are the full paths)
    signtool_path = tmp_path / "win_sdk/bin/83.0.1.0/x64/signtool.exe"
    assert win_sdk.signtool_exe == signtool_path


def test_winsdk_nonlatest_install_from_reg(
    mock_tools,
    mock_winreg,
    tmp_path,
    capsys,
    monkeypatch,
):
    """If a subsequent version of the SDK found in the registry is valid, the validator
    succeeds."""
    # Bypass using SDK in env vars
    mock_tools.os.environ.get.return_value = None

    # Turn on verbose logging
    mock_tools.logger.verbosity = LogLevel.DEBUG

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "85.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Set up an SDK installation without expected binaries
    WindowsSDK.SDK_VERSION = "85.0"
    invalid_sdk_path, invalid_sdk_ver = setup_winsdk_install(
        tmp_path / "invalid", "85.0.9", skip_bins=True
    )

    # Mock the SDK install to be used
    sdk_path, sdk_ver = setup_winsdk_install(tmp_path, "85.0.8")

    # Mock the registry queries via QueryValueEx
    mock_winreg.QueryValueEx.side_effect = [
        (str(invalid_sdk_path), mock_winreg.REG_SZ),
        (invalid_sdk_ver, mock_winreg.REG_SZ),
        (str(sdk_path), mock_winreg.REG_SZ),
        (sdk_ver, mock_winreg.REG_SZ),
    ]

    # Verify the install
    win_sdk = WindowsSDK.verify(mock_tools)

    # Confirm invalid SDK was evaluated
    expected_output = (
        "\n"
        "[Windows SDK] Finding Suitable Installation...\n"
        f"Evaluating Registry SDK version '85.0.9.0' at {tmp_path / 'invalid' / 'win_sdk'}\n"
        f"Evaluating Registry SDK version '85.0.8.0' at {tmp_path / 'win_sdk'}\n"
        f"Using Windows SDK v85.0.8.0 at {tmp_path / 'win_sdk'}\n"
    )
    assert capsys.readouterr().out == expected_output

    # The returned paths are as expected (and are the full paths)
    signtool_path = tmp_path / "win_sdk/bin/85.0.8.0/x64/signtool.exe"
    assert win_sdk.signtool_exe == signtool_path


@pytest.mark.parametrize(
    "reg_installs, additional_installs",
    [
        # One invalid registry install; no additional installs
        [[("invalid_1", "85.0.1")], []],
        # One invalid registry install with missing SDK version; no additional installs
        [[("invalid_1", "")], []],
        # One invalid registry install but directory key lookup fails; no additional installs
        [[("invalid_1", "85.0.1"), (FileNotFoundError, "")], []],
        # One invalid registry install but version key lookup fails; no additional installs
        [[("invalid_1", "85.0.1"), ("invalid_1", FileNotFoundError)], []],
        # Multiple invalid registry installs; no additional installs
        [[("invalid_1", "85.0.1"), ("invalid_2", "85.0.2")], []],
        [[("invalid_1", "85.0.0"), ("invalid_2", "86.0.2")], []],
        # One invalid registry install; one additional invalid install
        [[("invalid_2", "85.0.1")], [("invalid_2", "85.0.2")]],
        # One invalid registry install; multiple additional invalid installs
        [[("invalid_3", "85.0.1")], [("invalid_3", "85.0.3"), ("invalid_3", "85.0.2")]],
        [[("invalid_3", "")], [("invalid_3", "85.0.3"), ("invalid_3", "85.0.2")]],
        # Multiple invalid registry installs; one additional install
        [[("invalid_4", "85.0.1"), ("invalid_5", "85.0.3")], [("invalid_5", "85.0.4")]],
        [
            [("invalid_4", FileNotFoundError), ("invalid_5", "85.0.3")],
            [("invalid_5", "85.0.4")],
        ],
        [[(FileNotFoundError, ""), ("invalid_5", "85.0.3")], [("invalid_5", "85.0.4")]],
    ],
)
def test_winsdk_invalid_install_from_reg(
    mock_tools,
    mock_winreg,
    reg_installs,
    additional_installs,
    tmp_path,
    capsys,
    monkeypatch,
):
    """If none of the SDKs found in the registry are valid, the validator fails."""
    # Bypass using SDK in env vars
    mock_tools.os.environ.get.return_value = None

    # Bypass using default SDK dirs
    WindowsSDK.DEFAULT_SDK_DIRS = []

    # Turn on verbose logging
    mock_tools.logger.verbosity = LogLevel.DEBUG

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "85.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 1)

    # Set up SDK installations without expected binaries
    reg_queries = []
    for subdir, version in reg_installs:
        reg_queries.extend(
            setup_winsdk_install(tmp_path / subdir, version, skip_bins=True)
            if FileNotFoundError not in {subdir, version}
            else [subdir, version]
        )
    for subdir, version in additional_installs:
        setup_winsdk_install(tmp_path / subdir, version, skip_bins=True)
    # Add enough exceptions to run out any remaining registry reads
    reg_queries.extend([FileNotFoundError] * 12)

    # Mock the registry queries for install dir and version
    mock_winreg.QueryValueEx.side_effect = [
        (q, mock_winreg.REG_SZ) if q is not FileNotFoundError else FileNotFoundError
        for q in reg_queries
    ]

    # Verify the install
    error_text = (
        "Unable to locate a suitable Windows SDK v85.0 installation.\n"
        "\n"
        "Ensure at least v85.0.1.0 is installed and the components below are included:\n"
        "\n"
        "    * Windows SDK Signing Tools for Desktop Apps\n"
        "    * Windows SDK for UWP Managed Apps\n"
        "\n"
        "See https://developer.microsoft.com/en-us/windows/downloads/windows-sdk/ to install the SDK.\n"
    )
    with pytest.raises(BriefcaseCommandError, match=re.escape(error_text)):
        WindowsSDK.verify(mock_tools)

    # Confirm invalid SDK was evaluated
    expected_output = "\n[Windows SDK] Finding Suitable Installation...\n"
    for subdir, version in [t for t in reg_installs if FileNotFoundError not in t]:
        if subdir and version:
            sdk_path = tmp_path / subdir / "win_sdk"
            expected_output += (
                f"Evaluating Registry SDK version '{version}.0' at {sdk_path}\n"
            )
    for subdir, version in additional_installs:
        sdk_path = tmp_path / subdir / "win_sdk"
        expected_output += (
            f"Evaluating Registry SDK Bin version '{version}.0' at {sdk_path}\n"
        )
    assert capsys.readouterr().out == expected_output


def test_winsdk_valid_install_from_default_dir(
    mock_tools,
    mock_winreg,
    tmp_path,
    capsys,
    monkeypatch,
):
    """If an SDKs in a default directory is valid, the validator succeeds."""
    # Bypass using SDK in env vars
    mock_tools.os.environ.get.return_value = None

    # Turn on verbose logging
    mock_tools.logger.verbosity = LogLevel.DEBUG

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "86.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Set up SDK installations
    sdk_dir, _ = setup_winsdk_install(tmp_path, "86.0.8")
    invalid_sdk_path, _ = setup_winsdk_install(
        tmp_path / "invalid", "86.0.7", skip_bins=True
    )
    WindowsSDK.DEFAULT_SDK_DIRS = [invalid_sdk_path, sdk_dir]

    # Verify the install
    win_sdk = WindowsSDK.verify(mock_tools)

    # Confirm invalid SDK was evaluated
    expected_output = (
        "\n"
        "[Windows SDK] Finding Suitable Installation...\n"
        f"Evaluating Default Bin SDK version '86.0.7.0' at {tmp_path / 'invalid' / 'win_sdk'}\n"
        f"Evaluating Default Bin SDK version '86.0.8.0' at {tmp_path / 'win_sdk'}\n"
        f"Using Windows SDK v86.0.8.0 at {tmp_path / 'win_sdk'}\n"
    )
    assert capsys.readouterr().out == expected_output

    # The returned paths are as expected (and are the full paths)
    signtool_path = tmp_path / "win_sdk/bin/86.0.8.0/x64/signtool.exe"
    assert win_sdk.signtool_exe == signtool_path


def test_winsdk_invalid_install_from_default_dir(
    mock_tools,
    mock_winreg,
    tmp_path,
    capsys,
    monkeypatch,
):
    """If none of the SDKs in the default directories are valid, the validator fails."""
    # Bypass using SDK in env vars
    mock_tools.os.environ.get.return_value = None

    # Turn on verbose logging
    mock_tools.logger.verbosity = LogLevel.DEBUG

    # Patch target SDK version
    monkeypatch.setattr(WindowsSDK, "SDK_VERSION", "87.0")
    monkeypatch.setattr(WindowsSDK, "SDK_MIN_VERSION", 0)

    # Set up SDK installations
    invalid_sdk_path_1, _ = setup_winsdk_install(
        tmp_path / "invalid_1", "87.0.7", skip_bins=True
    )
    invalid_sdk_path_2, _ = setup_winsdk_install(
        tmp_path / "invalid_2", "87.0.8", skip_bins=True
    )
    WindowsSDK.DEFAULT_SDK_DIRS = [
        tmp_path / "nonexistent_dir",
        invalid_sdk_path_1,
        invalid_sdk_path_2,
    ]

    # Verify the install
    with pytest.raises(
        BriefcaseCommandError,
        match="Unable to locate a suitable Windows SDK v87.0 installation.",
    ):
        WindowsSDK.verify(mock_tools)

    # Confirm invalid SDK was evaluated
    expected_output = (
        "\n"
        "[Windows SDK] Finding Suitable Installation...\n"
        f"Evaluating Default Bin SDK version '87.0.7.0' at {tmp_path / 'invalid_1' / 'win_sdk'}\n"
        f"Evaluating Default Bin SDK version '87.0.8.0' at {tmp_path / 'invalid_2' / 'win_sdk'}\n"
    )
    assert capsys.readouterr().out == expected_output



from tempfile import TemporaryDirectory

import pytest

import briefcase.commands.convert
from briefcase.commands import ConvertCommand
from briefcase.commands.base import full_options
from briefcase.console import Console, Log
from tests.utils import DummyConsole


class DummyConvertCommand(ConvertCommand):
    """A dummy convert command that doesn't actually do anything.

    It only serves to track which actions would be performed.
    """

    description = "Dummy convert command"

    def __init__(self, *args, **kwargs):
        kwargs.setdefault("logger", Log())
        kwargs.setdefault("console", Console())
        super().__init__(*args, apps={}, **kwargs)

        self.actions = []
        self.tools.input = DummyConsole()

    def verify_host(self):
        super().verify_host()
        self.actions.append(("verify-host",))

    def verify_tools(self):
        super().verify_tools()
        self.actions.append(("verify-tools",))

    def finalize_app_config(self, app):
        super().finalize_app_config(app=app)
        self.actions.append(("finalize-app-config", app))

    def verify_app_tools(self, app):
        super().verify_app_tools(app=app)
        self.actions.append(("verify-app-tools", app.app_name))

    def validate_pyproject_file(self):
        super().validate_pyproject_file()
        self.actions.append(("validate-pyproject-file",))

    def validate_not_empty_project(self):
        super().validate_not_empty_project()
        self.actions.append(("validate-not-empty-project",))

    def convert_app(self, **kwargs):
        self.actions.append(("new", kwargs))
        return full_options({"new_state": "done"}, kwargs)


@pytest.fixture
def convert_command(tmp_path):
    (tmp_path / "project").mkdir()
    return DummyConvertCommand(base_path=tmp_path / "project")


@pytest.fixture
def patch_tempdir(monkeypatch):
    tmpdir = TemporaryDirectory()
    monkeypatch.setattr(
        briefcase.commands.convert, "TemporaryDirectory", lambda: tmpdir
    )
    return tmpdir



import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_no_pyproject_file(convert_command):
    """If there is no pyproject.toml file, a BriefcaseCommandError is raised."""
    with pytest.raises(BriefcaseCommandError):
        convert_command.validate_pyproject_file()


def test_briefcase_field_present(convert_command):
    """If the project is already set up for Briefcase, a BriefcaseCommandError is
    raised."""
    (convert_command.base_path / "pyproject.toml").write_text(
        "[tool.briefcase]\n", encoding="utf-8"
    )
    with pytest.raises(BriefcaseCommandError):
        convert_command.validate_pyproject_file()


def test_no_briefcase_field(convert_command):
    """If the project has a pyproject.toml file and is not already set up for Briefcase,
    no error is raised."""
    (convert_command.base_path / "pyproject.toml").write_text(
        "[project]\n", encoding="utf-8"
    )
    convert_command.validate_pyproject_file()



import pytest

from briefcase.integrations.linuxdeploy import LinuxDeployLocalFilePlugin


@pytest.fixture
def linuxdeploy_local_file_plugin(mock_tools, tmp_path):
    return LinuxDeployLocalFilePlugin(
        mock_tools,
        plugin_path=tmp_path / "path/to/linuxdeploy-plugin-custom.sh",
        bundle_path=tmp_path / "bundle",
    )


def test_file_path(linuxdeploy_local_file_plugin, tmp_path):
    """Local file plugins are kept in the bundle."""
    assert linuxdeploy_local_file_plugin.file_path == tmp_path / "bundle"


def test_file_name(linuxdeploy_local_file_plugin):
    """Local file plugin filename is as provided."""
    assert linuxdeploy_local_file_plugin.file_name == "linuxdeploy-plugin-custom.sh"


def test_plugin_id(linuxdeploy_local_file_plugin):
    """The custom plugin ID can be determined from the filename."""
    assert linuxdeploy_local_file_plugin.plugin_id == "custom"


def test_download_url(linuxdeploy_local_file_plugin):
    """Local file plugins don't have a download URL."""
    # Local file plugins don't have download URL.
    with pytest.raises(RuntimeError):
        _ = linuxdeploy_local_file_plugin.download_url



import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_publish(publish_command, first_app, second_app):
    """If there are multiple apps, publish all of them."""
    # Add two apps
    publish_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = publish_command.parse_options([])

    # Run the publish command
    publish_command(**options)

    # The right sequence of things will be done
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Publish the first app to s3
        ("publish", "first", "s3", {}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Publish the second app to s3
        ("publish", "second", "s3", {"publish_state": "first"}),
    ]


def test_publish_alternative_channel(publish_command, first_app, second_app):
    """Apps can be published to an alternate channel."""
    # Add two apps
    publish_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = publish_command.parse_options(["-c", "alternative"])

    # Run the publish command
    publish_command(**options)

    # The right sequence of things will be done
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Publish the first app to the alternative channel
        ("publish", "first", "alternative", {}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Publish the second app to the alternative channel
        ("publish", "second", "alternative", {"publish_state": "first"}),
    ]


def test_non_existent(publish_command, first_app_config, second_app):
    """Requesting a publish of a non-existent app raises an error."""
    # Add two apps; use the "config only" version of the first app.
    publish_command.apps = {
        "first": first_app_config,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = publish_command.parse_options([])

    # Invoking the publish command raises an error
    with pytest.raises(BriefcaseCommandError):
        publish_command(**options)

    # Only verification will be performed
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
    ]


def test_unbuilt(publish_command, first_app_unbuilt, second_app):
    """Requesting a publish of an app that has been created, but not built, raises an
    error."""
    # Add two apps; use the "config only" version of the first app.
    publish_command.apps = {
        "first": first_app_unbuilt,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = publish_command.parse_options([])

    # Invoking the publish command raises an error
    with pytest.raises(BriefcaseCommandError):
        publish_command(**options)

    # Only verification will be performed
    assert publish_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
    ]



import os
import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.docker import DockerAppContext


@pytest.mark.parametrize("extra_args", [[], ["--option-one", "--option-two"]])
def test_prepare(mock_tools, my_app, extra_args, tmp_path, sub_stream_kw):
    """The Docker environment can be prepared."""
    mock_docker_app_context = DockerAppContext(mock_tools, my_app)
    mock_docker_app_context.prepare(
        image_tag="briefcase/com.example.myapp:py3.X",
        dockerfile_path=tmp_path / "bundle/Dockerfile",
        app_base_path=tmp_path / "base",
        host_bundle_path=tmp_path / "bundle",
        host_data_path=tmp_path / "briefcase",
        python_version="3.X",
        extra_build_args=extra_args,
    )

    mock_docker_app_context.tools.subprocess._subprocess.Popen.assert_called_with(
        [
            "docker",
            "buildx",
            "build",
            "--load",
            "--progress",
            "plain",
            "--tag",
            "briefcase/com.example.myapp:py3.X",
            "--file",
            os.fsdecode(tmp_path / "bundle/Dockerfile"),
            "--build-arg",
            "SYSTEM_REQUIRES=things==1.2 stuff>=3.4 runtime_things==1.42 stuff>=3.4",
            "--build-arg",
            "HOST_UID=37",
            "--build-arg",
            "HOST_GID=42",
            os.fsdecode(tmp_path / "base/path/to/src"),
        ]
        + extra_args,
        **sub_stream_kw,
    )

    assert mock_docker_app_context.app_base_path == tmp_path / "base"
    assert mock_docker_app_context.host_bundle_path == tmp_path / "bundle"
    assert mock_docker_app_context.host_data_path == tmp_path / "briefcase"
    assert mock_docker_app_context.image_tag == "briefcase/com.example.myapp:py3.X"
    assert mock_docker_app_context.python_version == "3.X"


def test_prepare_failure(mock_docker_app_context, tmp_path, sub_stream_kw):
    """If the Docker environment can't be prepared, an error is raised."""
    # Mock a failure in docker build.
    mock_docker_app_context.tools.subprocess._subprocess.Popen.side_effect = (
        subprocess.CalledProcessError(returncode=1, cmd="docker build")
    )

    with pytest.raises(BriefcaseCommandError):
        mock_docker_app_context.prepare(
            image_tag="briefcase/com.example.myapp:py3.X",
            dockerfile_path=tmp_path / "bundle/Dockerfile",
            app_base_path=tmp_path / "base",
            host_bundle_path=tmp_path / "bundle",
            host_data_path=tmp_path / "briefcase",
            python_version="3.X",
            extra_build_args=["--option-one", "--option-two"],
        )

    mock_docker_app_context.tools.subprocess._subprocess.Popen.assert_called_with(
        [
            "docker",
            "buildx",
            "build",
            "--load",
            "--progress",
            "plain",
            "--tag",
            "briefcase/com.example.myapp:py3.X",
            "--file",
            os.fsdecode(tmp_path / "bundle/Dockerfile"),
            "--build-arg",
            "SYSTEM_REQUIRES=things==1.2 stuff>=3.4 runtime_things==1.42 stuff>=3.4",
            "--build-arg",
            "HOST_UID=37",
            "--build-arg",
            "HOST_GID=42",
            os.fsdecode(tmp_path / "base/path/to/src"),
            "--option-one",
            "--option-two",
        ],
        **sub_stream_kw,
    )



import pytest

from briefcase.exceptions import BriefcaseCommandError, UnsupportedHostError
from briefcase.integrations.linuxdeploy import LinuxDeployLocalFilePlugin

from .utils import create_mock_appimage


@pytest.mark.parametrize("host_os", ["Windows", "wonky"])
def test_unsupported_os(mock_tools, host_os):
    """When host OS is not supported, an error is raised."""
    mock_tools.host_os = host_os

    with pytest.raises(
        UnsupportedHostError,
        match=f"{LinuxDeployLocalFilePlugin.name} is not supported on {host_os}",
    ):
        LinuxDeployLocalFilePlugin.verify(mock_tools)


def test_verify(mock_tools, tmp_path):
    """Local file plugins are installed by copying."""
    plugin_path = tmp_path / "path/to/linuxdeploy-plugin-custom.sh"
    create_mock_appimage(plugin_path)

    LinuxDeployLocalFilePlugin.verify(
        mock_tools,
        plugin_path=plugin_path,
        bundle_path=tmp_path / "bundle",
    )

    # The plugin is copied into place
    assert (tmp_path / "bundle/linuxdeploy-plugin-custom.sh").exists()


def test_bad_path(mock_tools, tmp_path):
    """If the plugin file path is invalid, an error is raised."""
    plugin_path = tmp_path / "path/to/linuxdeploy-plugin-custom.sh"

    with pytest.raises(
        BriefcaseCommandError,
        match=r"Could not locate linuxdeploy plugin ",
    ):
        LinuxDeployLocalFilePlugin.verify(
            mock_tools,
            plugin_path=plugin_path,
            bundle_path=tmp_path / "bundle",
        )


def test_non_plugin(mock_tools, tmp_path):
    """If the plugin file path exists, but the filename doesn't match the pattern of a
    linuxdeploy plugin, an error is raised."""
    plugin_path = tmp_path / "path/to/not-a-plugin.exe"
    create_mock_appimage(plugin_path)

    with pytest.raises(
        BriefcaseCommandError,
        match=r"not-a-plugin.exe is not a linuxdeploy plugin",
    ):
        LinuxDeployLocalFilePlugin.verify(
            mock_tools,
            plugin_path=plugin_path,
            bundle_path=tmp_path / "bundle",
        )



import sys
from pathlib import Path, PurePosixPath

import pytest


@pytest.mark.usefixtures("mock_docker")
def test_dockerize_simple_call(mock_tools):
    """Simple Docker command is returned."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
    )

    assert args == dict(
        args=["docker", "run", "--rm", "best-image", "hello", "world"],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
def test_dockerize_interactive(mock_tools):
    """Docker command includes interactive session."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
        interactive=True,
    )

    assert args == dict(
        args=["docker", "run", "--rm", "-it", "best-image", "hello", "world"],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
def test_dockerize_mounts(mock_tools):
    """Docker command includes bind mounts."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
        mounts=[
            ("/source/path_one", "/target/path_one"),
            ("/source/path_two", "/target/path_two"),
        ],
    )

    assert args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--volume",
            "/source/path_one:/target/path_one:z",
            "--volume",
            "/source/path_two:/target/path_two:z",
            "best-image",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
@pytest.mark.usefixtures("mock_docker")
@pytest.mark.parametrize("cwd", ["/my/cwd", Path("/my/cwd"), PurePosixPath("/my/cwd")])
def test_dockerize_cwd(mock_tools, cwd):
    """Docker command includes working directory."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
        cwd=cwd,
    )

    assert args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--workdir",
            "/my/cwd",
            "best-image",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
@pytest.mark.usefixtures("mock_docker")
def test_dockerize_env(mock_tools):
    """Docker command includes environment variables."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
        env={
            "ENVVAR1": "value",
            "ENVVAR2": "value with space",
            "ENVVAR3": "/my/str/path",
            "ENVVAR4": Path("/my/path/path"),
            "ENVVAR5": PurePosixPath("/my/pure/path/path"),
        },
    )

    assert args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--env",
            "ENVVAR1=value",
            "--env",
            "ENVVAR2=value with space",
            "--env",
            "ENVVAR3=/my/str/path",
            "--env",
            "ENVVAR4=/my/path/path",
            "--env",
            "ENVVAR5=/my/pure/path/path",
            "best-image",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


@pytest.mark.usefixtures("mock_docker")
def test_dockerize_add_hosts(mock_tools):
    """Docker command includes hosts to map."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
        add_hosts=[
            ("host1.local", "1.1.1.1"),
            ("host2.local", "1.1.1.2"),
            ("host3.local", "example.com"),
        ],
    )

    assert args == dict(
        args=[
            "docker",
            "run",
            "--rm",
            "--add-host",
            "host1.local:1.1.1.1",
            "--add-host",
            "host2.local:1.1.1.2",
            "--add-host",
            "host3.local:example.com",
            "best-image",
            "hello",
            "world",
        ],
        env={"DOCKER_CLI_HINTS": "false"},
    )


DOCKERIZE_TEST_PARAMS = [
    (
        dict(
            interactive=True,
            mounts=[
                ("/source/path_one", "/target/path_one"),
                ("/source/path_two", "/target/path_two"),
            ],
            cwd="/my/cwd/inside",
            env={
                "ENVVAR1": "value",
                "ENVVAR2": "value with space",
                "ENVVAR3": "/my/str/path",
                "ENVVAR4": Path("/my/path/path"),
                "ENVVAR5": PurePosixPath("/my/pure/path/path"),
            },
            add_hosts=[
                ("host1.local", "1.1.1.1"),
                ("host2.local", "1.1.1.2"),
                ("host3.local", "example.com"),
            ],
        ),
        dict(
            args=[
                "docker",
                "run",
                "--rm",
                "-it",
                "--volume",
                "/source/path_one:/target/path_one:z",
                "--volume",
                "/source/path_two:/target/path_two:z",
                "--env",
                "ENVVAR1=value",
                "--env",
                "ENVVAR2=value with space",
                "--env",
                "ENVVAR3=/my/str/path",
                "--env",
                "ENVVAR4=/my/path/path",
                "--env",
                "ENVVAR5=/my/pure/path/path",
                "--workdir",
                "/my/cwd/inside",
                "--add-host",
                "host1.local:1.1.1.1",
                "--add-host",
                "host2.local:1.1.1.2",
                "--add-host",
                "host3.local:example.com",
                "best-image",
                "hello",
                "world",
            ],
            env={"DOCKER_CLI_HINTS": "false"},
        ),
    )
]


@pytest.mark.skipif(
    sys.platform == "win32", reason="Windows paths aren't converted in Docker context"
)
@pytest.mark.parametrize("in_kwargs, out_args_kwargs", DOCKERIZE_TEST_PARAMS)
@pytest.mark.usefixtures("mock_docker")
def test_dockerized_complex_call(mock_tools, in_kwargs, out_args_kwargs):
    """Docker command includes combination of options."""
    args = mock_tools.docker.dockerize_args(
        ["hello", "world"],
        image_tag="best-image",
        **in_kwargs,
    )

    assert args == out_args_kwargs



import pytest


@pytest.mark.parametrize(
    "bundle, candidate",
    [
        ("com.example", "example.com"),
        ("com.example.more", "more.example.com"),
    ],
)
def test_make_app_name(new_command, bundle, candidate):
    """A bundle can be converted into a domain name."""
    app_name = new_command.make_domain(bundle)
    assert app_name == candidate



import os
import sys
from unittest.mock import MagicMock

import pytest

from briefcase.integrations.linuxdeploy import (
    LinuxDeployGtkPlugin,
    LinuxDeployLocalFilePlugin,
    LinuxDeployQtPlugin,
    LinuxDeployURLPlugin,
)

from .utils import (
    create_mock_appimage,
    side_effect_create_mock_appimage,
    side_effect_create_mock_tool,
)


def test_no_plugins(linuxdeploy, mock_tools, tmp_path):
    """If there are no plugins, verify is a no-op."""

    plugins = linuxdeploy.verify_plugins([], bundle_path=tmp_path / "bundle")

    mock_tools.file.download.assert_not_called()

    assert plugins == {}


def test_gtk_plugin(linuxdeploy, mock_tools, tmp_path):
    """The GTK plugin can be verified."""

    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_tool(
        tmp_path / "tools/linuxdeploy_plugins/gtk/linuxdeploy-plugin-gtk.sh"
    )

    plugins = linuxdeploy.verify_plugins(["gtk"], bundle_path=tmp_path / "bundle")

    assert plugins.keys() == {"gtk"}
    assert isinstance(plugins["gtk"], LinuxDeployGtkPlugin)

    mock_tools.file.download.assert_called_with(
        url="https://raw.githubusercontent.com/linuxdeploy/linuxdeploy-plugin-gtk/master/linuxdeploy-plugin-gtk.sh",
        download_path=tmp_path / "tools/linuxdeploy_plugins/gtk",
        role="linuxdeploy GTK plugin",
    )


def test_qt_plugin(linuxdeploy, mock_tools, tmp_path):
    """The Qt plugin can be verified."""

    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_appimage(
        tmp_path
        / "tools"
        / "linuxdeploy_plugins"
        / "qt"
        / "linuxdeploy-plugin-qt-i386.AppImage"
    )

    plugins = linuxdeploy.verify_plugins(["qt"], bundle_path=tmp_path / "bundle")

    assert plugins.keys() == {"qt"}
    assert isinstance(plugins["qt"], LinuxDeployQtPlugin)

    mock_tools.file.download.assert_called_with(
        url=(
            "https://github.com/linuxdeploy/linuxdeploy-plugin-qt/"
            "releases/download/continuous/linuxdeploy-plugin-qt-i386.AppImage"
        ),
        download_path=tmp_path / "tools/linuxdeploy_plugins/qt",
        role="linuxdeploy Qt plugin",
    )


def test_custom_url_plugin(linuxdeploy, mock_tools, tmp_path):
    """A Custom URL plugin can be verified."""

    # Mock a successful download
    mock_tools.file.download.side_effect = side_effect_create_mock_appimage(
        tmp_path
        / "tools"
        / "linuxdeploy_plugins"
        / "sometool"
        / "bbf1b5dc4c3d2069dc5b916f73e9d6f5ad24603576298509367878e393c6f8f5"
        / "linuxdeploy-plugin-sometool-i386.AppImage"
    )

    plugins = linuxdeploy.verify_plugins(
        ["https://example.com/path/to/linuxdeploy-plugin-sometool-i386.AppImage"],
        bundle_path=tmp_path / "bundle",
    )

    assert plugins.keys() == {"sometool"}
    assert isinstance(plugins["sometool"], LinuxDeployURLPlugin)

    mock_tools.file.download.assert_called_with(
        url="https://example.com/path/to/linuxdeploy-plugin-sometool-i386.AppImage",
        download_path=tmp_path
        / "tools"
        / "linuxdeploy_plugins"
        / "sometool"
        / "bbf1b5dc4c3d2069dc5b916f73e9d6f5ad24603576298509367878e393c6f8f5",
        role="user-provided linuxdeploy plugin from URL",
    )


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows paths can't be passed to linuxdeploy",
)
def test_custom_local_file_plugin(linuxdeploy, mock_tools, tmp_path):
    """A Custom local file plugin can be verified."""

    # Create a local file
    plugin_path = tmp_path / "path/to/linuxdeploy-plugin-sometool-i386.AppImage"
    create_mock_appimage(plugin_path)

    plugins = linuxdeploy.verify_plugins(
        [os.fsdecode(plugin_path)],
        bundle_path=tmp_path / "bundle",
    )

    assert plugins.keys() == {"sometool"}
    assert isinstance(plugins["sometool"], LinuxDeployLocalFilePlugin)

    # No download happened
    mock_tools.file.download.assert_not_called()
    # But a copy happened
    assert (tmp_path / "bundle/linuxdeploy-plugin-sometool-i386.AppImage").exists()


@pytest.mark.parametrize(
    "plugin_declaration, expected_env",
    [
        # No environment variables
        (
            "gtk",
            {},
        ),
        # Single environment variable
        (
            "FOO=bar gtk",
            {"FOO": "bar"},
        ),
        # Multiple environment variables
        (
            "FOO=bar PORK=ham gtk",
            {"FOO": "bar", "PORK": "ham"},
        ),
        # Space in var escaped with quotes
        (
            "FOO=bar PORK='serrano ham' gtk",
            {"FOO": "bar", "PORK": "serrano ham"},
        ),
        # Space escaped with baskslash
        (
            "FOO=bar PORK=serrano\\ ham gtk",
            {"FOO": "bar", "PORK": "serrano ham"},
        ),
        # Key-only definition
        (
            "FOO=bar PORK gtk",
            {"FOO": "bar", "PORK": ""},
        ),
    ],
)
def test_plugin_env(
    linuxdeploy,
    tmp_path,
    plugin_declaration,
    expected_env,
):
    """All linuxdeploy installs are managed."""
    linuxdeploy.plugins["gtk"].is_elf_file = MagicMock(return_value=False)

    plugins = linuxdeploy.verify_plugins(
        [plugin_declaration],
        bundle_path=tmp_path / "bundle",
    )

    assert plugins.keys() == {"gtk"}
    assert isinstance(plugins["gtk"], LinuxDeployGtkPlugin)
    assert plugins["gtk"].env == expected_env


@pytest.mark.skipif(
    sys.platform == "win32",
    reason="Windows paths can't be passed to linuxdeploy",
)
def test_complex_plugin_config(linuxdeploy, mock_tools, tmp_path):
    """A complex plugin configuration can be verified."""
    # Define multiple plugins, of different types, each with different environments

    # Three tools are obtained by downloading.
    # We don't want the side effects to occur until the function is invoked;
    # so we need to wrap the side effect callables in another callable.
    def mock_downloads(url, download_path, role):
        if "linuxdeploy_plugins/gtk" in str(download_path):
            return side_effect_create_mock_tool(
                tmp_path
                / "tools"
                / "linuxdeploy_plugins"
                / "gtk"
                / "linuxdeploy-plugin-gtk.sh"
            )(url, download_path, role)
        elif "linuxdeploy_plugins/qt" in str(download_path):
            return side_effect_create_mock_appimage(
                tmp_path
                / "tools"
                / "linuxdeploy_plugins"
                / "qt"
                / "linuxdeploy-plugin-qt-i386.AppImage"
            )(url, download_path, role)
        elif "linuxdeploy_plugins/network" in str(download_path):
            return side_effect_create_mock_tool(
                tmp_path
                / "tools"
                / "linuxdeploy_plugins"
                / "network"
                / "f93c7e6d04425e4ed7e533655d9898c4984f104711f4ffd3b3966cfc92500b2d"
                / "linuxdeploy-plugin-network.sh"
            )(url, download_path, role)
        else:
            raise Exception("Unexpected download")

    mock_tools.file.download.side_effect = mock_downloads

    # Local file tool is a local file.
    local_plugin_path = tmp_path / "path/to/linuxdeploy-plugin-sometool-i386.AppImage"
    create_mock_appimage(local_plugin_path)

    # Verify all the plugins
    plugins = linuxdeploy.verify_plugins(
        [
            "DEPLOY_GTK_VERSION=3 gtk",
            "qt",
            f"QUALITY='really nice' {local_plugin_path}",
            "FOO=bar PORK=ham https://example.com/path/to/linuxdeploy-plugin-network.sh",
        ],
        bundle_path=tmp_path / "bundle",
    )

    assert plugins.keys() == {"gtk", "qt", "sometool", "network"}

    # GTK plugin is as expected
    assert isinstance(plugins["gtk"], LinuxDeployGtkPlugin)
    assert plugins["gtk"].env == {"DEPLOY_GTK_VERSION": "3"}

    # Qt plugin is as expected
    assert isinstance(plugins["qt"], LinuxDeployQtPlugin)
    assert plugins["qt"].env == {}

    # Local file plugin is as expected
    assert isinstance(plugins["sometool"], LinuxDeployLocalFilePlugin)
    assert plugins["sometool"].env == {"QUALITY": "really nice"}
    assert (tmp_path / "bundle/linuxdeploy-plugin-sometool-i386.AppImage").exists()

    # URL plugin is as expected
    assert isinstance(plugins["network"], LinuxDeployURLPlugin)
    assert plugins["network"].env == {"FOO": "bar", "PORK": "ham"}



import subprocess

import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.integrations.xcode import XcodeCliTools


def test_license_accepted(capsys, mock_tools):
    """If the Xcode license has been accepted, pass without comment."""
    # Check passes without an error...
    XcodeCliTools.confirm_xcode_license_accepted(mock_tools)

    # ... clang was invoked ...
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["/usr/bin/clang", "--version"]
    )

    # ... and the user is none the wiser
    out = capsys.readouterr().out
    assert len(out) == 0


def test_unknown_error(capsys, mock_tools):
    """If an unexpected problem occurred accepting the license, warn the user."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["/usr/bin/clang", "--version"],
        returncode=1,
    )

    # Check passes without an error...
    XcodeCliTools.confirm_xcode_license_accepted(mock_tools)

    # ... clang was invoked ...
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["/usr/bin/clang", "--version"],
    )

    # ...but stdout contains a warning
    out = capsys.readouterr().out
    assert "************" in out


def test_accept_license(mock_tools):
    """If the user accepts the license, continue without error."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["/usr/bin/clang", "--version"],
        returncode=69,
    )

    # Check passes without an error...
    XcodeCliTools.confirm_xcode_license_accepted(mock_tools)

    # ... clang *and* xcodebuild were invoked ...
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["/usr/bin/clang", "--version"],
    )
    mock_tools.subprocess.run.assert_called_once_with(
        ["sudo", "xcodebuild", "-license"],
        check=True,
        stream_output=False,
    )


def test_sudo_fail(mock_tools):
    """If the sudo call fails, an exception is raised."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["/usr/bin/clang", "--version"],
        returncode=69,
    )
    mock_tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["sudo", "xcodebuild", "-license"],
        returncode=1,
    )

    # Check raises an error:
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Briefcase was unable to run the Xcode licensing tool.",
    ):
        XcodeCliTools.confirm_xcode_license_accepted(mock_tools)

    # ... clang *and* xcodebuild were invoked ...
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["/usr/bin/clang", "--version"],
    )
    mock_tools.subprocess.run.assert_called_once_with(
        ["sudo", "xcodebuild", "-license"],
        check=True,
        stream_output=False,
    )


def test_license_not_accepted(mock_tools):
    """If the sudo call fails, an exception is raised."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["/usr/bin/clang", "--version"],
        returncode=69,
    )
    mock_tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["sudo", "xcodebuild", "-license"],
        returncode=69,
    )

    # Check raises an error:
    with pytest.raises(
        BriefcaseCommandError, match=r"Xcode license has not been accepted."
    ):
        XcodeCliTools.confirm_xcode_license_accepted(mock_tools)

    # ... clang *and* xcodebuild were invoked ...
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["/usr/bin/clang", "--version"],
    )
    mock_tools.subprocess.run.assert_called_once_with(
        ["sudo", "xcodebuild", "-license"],
        check=True,
        stream_output=False,
    )


def test_license_status_unknown(capsys, mock_tools):
    """If we get an unusual response from the license, warn but continue."""
    mock_tools.subprocess.check_output.side_effect = subprocess.CalledProcessError(
        cmd=["/usr/bin/clang", "--version"],
        returncode=69,
    )
    mock_tools.subprocess.run.side_effect = subprocess.CalledProcessError(
        cmd=["sudo", "xcodebuild", "-license"],
        returncode=42,
    )

    # Check passes without error...
    XcodeCliTools.confirm_xcode_license_accepted(mock_tools)

    # ... clang *and* xcodebuild were invoked ...
    mock_tools.subprocess.check_output.assert_called_once_with(
        ["/usr/bin/clang", "--version"],
    )
    mock_tools.subprocess.run.assert_called_once_with(
        ["sudo", "xcodebuild", "-license"],
        check=True,
        stream_output=False,
    )

    # ...but stdout contains a warning
    out = capsys.readouterr().out
    assert "************" in out



from unittest.mock import MagicMock

import pytest

from briefcase.integrations.base import ToolCache
from briefcase.integrations.visualstudio import VisualStudio


@pytest.fixture
def visualstudio(mock_tools, tmp_path) -> VisualStudio:
    msbuild_path = tmp_path / "Visual Studio/MSBuild.exe"
    return VisualStudio(mock_tools, msbuild_path=msbuild_path)


def test_managed_install(visualstudio, tmp_path):
    """All Visual Studio installs are unmanaged."""
    assert not visualstudio.managed_install


def test_msbuild_path(visualstudio, tmp_path):
    """The MSBuild path is the one used for construction."""
    assert visualstudio.msbuild_path == tmp_path / "Visual Studio/MSBuild.exe"


def test_install_metadata(tmp_path):
    """Install metadata can be provided and retrieved."""
    msbuild_path = tmp_path / "Visual Studio/MSBuild.exe"
    install_metadata = {
        "instanceId": "deadbeef",
        "installDate": "2022-07-14T10:42:37Z",
    }
    visualstudio = VisualStudio(
        MagicMock(spec_set=ToolCache),
        msbuild_path=msbuild_path,
        install_metadata=install_metadata,
    )

    assert visualstudio.install_metadata["instanceId"] == "deadbeef"
    assert visualstudio.install_metadata["installDate"] == "2022-07-14T10:42:37Z"


def test_no_install_metadata(visualstudio, tmp_path):
    """Install metadata is optional."""
    assert visualstudio.install_metadata is None



import pytest

from briefcase.exceptions import CorruptToolError
from briefcase.integrations.linuxdeploy import (
    ELF_PATCH_OFFSET,
    ELF_PATCH_ORIGINAL_BYTES,
    ELF_PATCH_PATCHED_BYTES,
)
from tests.integrations.linuxdeploy.utils import create_mock_appimage


def test_patch_linuxdeploy_elf_header_unpatched(linuxdeploy, tmp_path):
    """If the linuxdeploy tool/plugin is not patched, patch it."""
    appimage_path = tmp_path / "tools/linuxdeploy-i386.AppImage"

    # Mock an unpatched linuxdeploy AppImage
    pre_patch_header = create_mock_appimage(
        appimage_path=appimage_path, mock_appimage_kind="original"
    )

    # Create a linuxdeploy wrapper, then patch the elf header
    linuxdeploy.patch_elf_header()

    # Ensure the patch was applied.
    with open(appimage_path, "rb") as mock_appimage:
        mock_appimage.seek(ELF_PATCH_OFFSET)
        patched_header = mock_appimage.read(len(ELF_PATCH_PATCHED_BYTES))

    assert pre_patch_header == ELF_PATCH_ORIGINAL_BYTES
    assert patched_header == ELF_PATCH_PATCHED_BYTES


def test_patch_linuxdeploy_elf_header_already_patched(linuxdeploy, tmp_path):
    """If linuxdeploy is already patched, don't patch it."""
    appimage_path = tmp_path / "tools/linuxdeploy-i386.AppImage"

    # Mock a patched linuxdeploy AppImage
    pre_patch_header = create_mock_appimage(
        appimage_path=appimage_path, mock_appimage_kind="patched"
    )

    # Create a linuxdeploy wrapper, then patch the elf header
    linuxdeploy.patch_elf_header()

    # Ensure the patch was applied.
    with open(appimage_path, "rb") as mock_appimage:
        mock_appimage.seek(ELF_PATCH_OFFSET)
        patched_header = mock_appimage.read(len(ELF_PATCH_PATCHED_BYTES))

    assert pre_patch_header == ELF_PATCH_PATCHED_BYTES
    assert patched_header == ELF_PATCH_PATCHED_BYTES


def test_patch_linuxdeploy_elf_header_bad_appimage(linuxdeploy, tmp_path):
    """If linuxdeploy does not have a valid header, raise an error."""
    appimage_path = tmp_path / "tools/linuxdeploy-i386.AppImage"

    # Mock a bad linuxdeploy AppImage
    create_mock_appimage(appimage_path=appimage_path, mock_appimage_kind="corrupt")

    # Create a linuxdeploy wrapper, then patch the elf header
    with pytest.raises(CorruptToolError):
        linuxdeploy = linuxdeploy.patch_elf_header()


def test_patch_linuxdeploy_elf_header_empty_appimage(linuxdeploy, tmp_path):
    """If file is empty, raise an error."""
    appimage_path = tmp_path / "tools/linuxdeploy-i386.AppImage"

    # Mock a bad linuxdeploy AppImage
    create_mock_appimage(appimage_path=appimage_path, mock_appimage_kind="empty")

    # Create a linuxdeploy wrapper, then patch the elf header
    with pytest.raises(CorruptToolError):
        linuxdeploy.patch_elf_header()



import re

import pytest

from briefcase.console import Printer


@pytest.fixture
def printer():
    printer = Printer()
    # clear log; since Printer is a pseudo-singleton,
    # the log can contain existing entries.
    printer.export_log()
    return printer


def norm_sp(text: str, max_spaces: int = 100) -> str:
    """Normalize more than `max_spaces` spaces in text to `max_spaces` spaces.

    This is necessary until https://github.com/Textualize/rich/issues/2944 is resolved
    to ensure that tests succeed while running in `tox` on Windows.
    """
    return re.sub(rf"\s{{{max_spaces},}}", " " * max_spaces, text)


@pytest.mark.parametrize(
    "show, message, expected_console_output",
    [
        (True, "a line of output", "a line of output\n"),
        (False, "a line of output", ""),
    ],
)
def test_call(capsys, printer, message, show, expected_console_output):
    """Printer prints to console and log appropriately."""
    printer(message, show=show, stack_offset=1)
    assert capsys.readouterr().out == expected_console_output
    log = printer.export_log()
    assert len(log.splitlines()) == 1
    # The number of spaces in not consistent on Windows
    assert norm_sp(" " + message + " " * 139 + "console.py") in norm_sp(log)


def test_to_console(capsys, printer):
    """Printer prints only to console."""
    printer.to_console("a line of output")
    assert capsys.readouterr().out == "a line of output\n"
    assert printer.export_log() == ""


def test_to_log(capsys, printer):
    """Printer prints only to log."""
    printer.to_log("a line of output", stack_offset=1)
    assert capsys.readouterr().out == ""
    log = printer.export_log()
    assert len(log.splitlines()) == 1
    # The number of spaces in not consistent on Windows
    assert norm_sp(" a line of output" + " " * 139 + "console.py") in norm_sp(log)


def test_very_long_line(capsys, printer):
    """Very long lines are split."""
    printer.to_log("A very long line of output!! " * 6, stack_offset=1)
    assert capsys.readouterr().out == ""
    log = printer.export_log()
    assert len(log.splitlines()) == 2
    # The number of spaces in not consistent on Windows
    assert norm_sp(
        " " + "A very long line of output!! " * 5 + "A very    console.py:",
        max_spaces=3,
    ) in norm_sp(log, max_spaces=3)
    assert norm_sp(" long line of output!!" + " " * 148) in norm_sp(log)



# This file contains raw log lines, which are extremely long. Turn of the flake8 rule
# for line length.
# flake8: noqa: E501
import pytest

from briefcase.platforms.macOS.filters import XcodeBuildFilter


@pytest.mark.parametrize(
    "original, filtered",
    [
        # Nothing to filter
        (
            [
                "'Twas brillig, and the slithy toves",
                "Did gyre and gimble in the wabe;",
                "All mimsy were the borogoves,",
                "And the mome raths outgrabe.",
            ],
            [
                "'Twas brillig, and the slithy toves",
                "Did gyre and gimble in the wabe;",
                "All mimsy were the borogoves,",
                "And the mome raths outgrabe.",
            ],
        ),
        # Xcode 14: DTDKRemoteDeviceConnection warning about locked attached devices
        (
            [
                "'Twas brillig, and the slithy toves",
                '2023-09-27 08:38:11.865 xcodebuild[41087:25901835]  DTDKRemoteDeviceConnection: Failed to start remote service "com.apple.mobile.notification_proxy" on device. Error: Error Domain=com.apple.dtdevicekit Code=811 "Failed to start remote service "com.apple.mobile.notification_proxy" on device." UserInfo={NSUnderlyingError=0x10b8ec780 {Error Domain=com.apple.dt.MobileDeviceErrorDomain Code=-402653158 "The device is passcode protected." UserInfo={MobileDeviceErrorCode=(0xE800001A), com.apple.dtdevicekit.stacktrace=(',
                "        0   DTDeviceKitBase                     0x00000001288ff298 DTDKCreateNSErrorFromAMDErrorCode + 300",
                "        1   DTDeviceKitBase                     0x000000012890ba38 __63-[DTDKRemoteDeviceConnection startFirstServiceOf:unlockKeybag:]_block_invoke + 380",
                "        2   DTDeviceKitBase                     0x000000012890b248 __48-[DTDKRemoteDeviceConnection futureWithSession:]_block_invoke_4 + 28",
                "        3   DTDeviceKitBase                     0x0000000128901460 __DTDKExecuteInSession_block_invoke_2 + 68",
                "        4   DTDeviceKitBase                     0x0000000128900af0 __DTDKExecuteWithConnection_block_invoke_2 + 216",
                "        5   DTDeviceKitBase                     0x00000001289009e8 __DTDKExecuteWithConnection_block_invoke + 112",
                "        6   libdispatch.dylib                   0x00000001a81b4400 _dispatch_client_callout + 20",
                "        7   libdispatch.dylib                   0x00000001a81c397c _dispatch_lane_barrier_sync_invoke_and_complete + 56",
                "        8   DVTFoundation                       0x0000000100fa8014 DVTDispatchBarrierSync + 148",
                "        9   DVTFoundation                       0x0000000100f842b4 -[DVTDispatchLock performLockedBlock:] + 60",
                "        10  DTDeviceKitBase                     0x00000001289008e4 DTDKExecuteWithConnection + 200",
                "        11  DTDeviceKitBase                     0x00000001289012e4 DTDKExecuteInSession + 260",
                "        12  DTDeviceKitBase                     0x000000012890b028 __48-[DTDKRemoteDeviceConnection futureWithSession:]_block_invoke_2 + 204",
                "        13  DVTFoundation                       0x0000000100fa7330 __DVT_CALLING_CLIENT_BLOCK__ + 16",
                "        14  DVTFoundation                       0x0000000100fa7d58 __DVTDispatchAsync_block_invoke + 152",
                "        15  libdispatch.dylib                   0x00000001a81b2874 _dispatch_call_block_and_release + 32",
                "        16  libdispatch.dylib                   0x00000001a81b4400 _dispatch_client_callout + 20",
                "        17  libdispatch.dylib                   0x00000001a81bba88 _dispatch_lane_serial_drain + 668",
                "        18  libdispatch.dylib                   0x00000001a81bc62c _dispatch_lane_invoke + 436",
                "        19  libdispatch.dylib                   0x00000001a81c7244 _dispatch_workloop_worker_thread + 648",
                "        20  libsystem_pthread.dylib             0x00000001a8360074 _pthread_wqthread + 288",
                "        21  libsystem_pthread.dylib             0x00000001a835ed94 start_wqthread + 8",
                '), DVTRadarComponentKey=261622, NSLocalizedDescription=The device is passcode protected.}}, NSLocalizedRecoverySuggestion=Please check your connection to your "device., DVTRadarComponentKey=261622, NSLocalizedDescription=Failed to start remote service "com.apple.mobile.notification_proxy" on device.}',
                "Did gyre and gimble in the wabe;",
            ],
            [
                "'Twas brillig, and the slithy toves",
                "Did gyre and gimble in the wabe;",
            ],
        ),
        # Xcode 14: Additional locking-related message.
        (
            [
                "'Twas brillig, and the slithy toves",
                "2023-09-27 09:09:55.400 xcodebuild[44887:25948169] Failed to start service (com.apple.amfi.lockdown): 0xe800001a",
                "Did gyre and gimble in the wabe;",
            ],
            [
                "'Twas brillig, and the slithy toves",
                "Did gyre and gimble in the wabe;",
            ],
        ),
        # XCode 14: x86_64 "device enabled state" warning.
        (
            [
                "'Twas brillig, and the slithy toves",
                "2023-10-04 08:05:21.757 xcodebuild[46899:11335453] DVTCoreDeviceEnabledState: DVTCoreDeviceEnabledState_Disabled set via user default (DVTEnableCoreDevice=disabled)",
                "Did gyre and gimble in the wabe;",
            ],
            [
                "'Twas brillig, and the slithy toves",
                "Did gyre and gimble in the wabe;",
            ],
        ),
        # Xcode 15: DVTAssertions warning about createItemModels
        (
            [
                "'Twas brillig, and the slithy toves",
                "2023-09-26 14:35:45.775 xcodebuild[75877:23947967] [MT] DVTAssertions: Warning in /System/Volumes/Data/SWE/Apps/DT/BuildRoots/BuildRoot11/ActiveBuildRoot/Library/Caches/com.apple.xbs/Sources/IDEFrameworks/IDEFrameworks-22267/IDEFoundation/Provisioning/Capabilities Infrastructure/IDECapabilityQuerySelection.swift:103",
                "Details: createItemModels creation requirements should not create capability item model for a capability item model that already exists.",
                "Function: createItemModels(for:itemModelSource:)",
                "Thread:   <_NSMainThread: 0x11d60beb0>{number = 1, name = main}",
                "Please file a bug at https://feedbackassistant.apple.com with this warning message and any useful information you can provide.",
                "Did gyre and gimble in the wabe;",
            ],
            [
                "'Twas brillig, and the slithy toves",
                "Did gyre and gimble in the wabe;",
            ],
        ),
    ],
)
def test_filter(original, filtered):
    """The Xcode build output filters out ignorable warnings."""
    xcode_filter = XcodeBuildFilter()

    output = []
    for raw_line in original:
        for line in xcode_filter(raw_line):
            output.append(line)

    assert output == filtered



import contextlib
import datetime
import logging
from io import TextIOBase
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, call

import pytest
from rich.traceback import Trace

import briefcase
from briefcase.commands.dev import DevCommand
from briefcase.console import Console, Log, LogLevel, Printer, RichLoggingHandler
from briefcase.exceptions import BriefcaseError

TRACEBACK_HEADER = "Traceback (most recent call last)"
EXTRA_HEADER = "Extra information:"


@pytest.fixture
def mock_now(monkeypatch):
    """Monkeypatch the ``datetime.now`` inside ``briefcase.console``.

    When this fixture is used, the log filename for the test will be
    ``briefcase.2022_06_25-16_12_29.dev.log``, assuming the command is a DevCommand.
    """
    now = datetime.datetime(2022, 6, 25, 16, 12, 29)
    datetime_mock = MagicMock(wraps=datetime.datetime)
    datetime_mock.now.return_value = now
    monkeypatch.setattr(briefcase.console, "datetime", datetime_mock)
    return now


@pytest.fixture
def command(mock_now, tmp_path) -> DevCommand:
    """Provides a mocked DevCommand."""
    command = MagicMock(spec_set=DevCommand(Log(), Console()))
    command.base_path = tmp_path
    command.command = "dev"
    command.tools.os.environ = {}
    return command


@pytest.fixture
def logging_logger() -> logging.Logger:
    logging_logger = logging.getLogger("test_pkg")
    yield logging_logger
    # reset handlers since they are persistent
    logging_logger.handlers.clear()


@pytest.mark.parametrize(
    "verbosity, verbose_enabled, debug_enabled, deep_debug_enabled",
    [
        (-1, False, False, False),
        (0, False, False, False),
        (LogLevel.INFO, False, False, False),
        (1, True, False, False),
        (LogLevel.VERBOSE, True, False, False),
        (2, True, True, False),
        (LogLevel.DEBUG, True, True, False),
        (3, True, True, True),
        (LogLevel.DEEP_DEBUG, True, True, True),
        (4, True, True, True),
        (5, True, True, True),
    ],
)
def test_log_level(verbosity, verbose_enabled, debug_enabled, deep_debug_enabled):
    """Logging level is correct."""
    assert Log(verbosity=verbosity).is_verbose is verbose_enabled
    assert Log(verbosity=verbosity).is_debug is debug_enabled
    assert Log(verbosity=verbosity).is_deep_debug is deep_debug_enabled


def test_info_logging(capsys):
    """The info level logging only includes info logs."""
    logger = Log()

    logger.info("info")
    logger.verbose("verbose")
    logger.debug("debug")

    output = capsys.readouterr().out.splitlines()

    assert "info" in output
    assert "verbose" not in output
    assert "debug" not in output


def test_verbose_logging(capsys):
    """The verbose level logging includes info and verbose logs."""
    logger = Log(verbosity=LogLevel.VERBOSE)

    logger.info("info")
    logger.verbose("verbose")
    logger.debug("debug")

    output = capsys.readouterr().out.splitlines()

    assert "info" in output
    assert "verbose" in output
    assert "debug" not in output


def test_debug_logging(capsys):
    """The debug level logging includes info, verbose and debug logs."""
    logger = Log(verbosity=LogLevel.DEBUG)

    logger.info("info")
    logger.verbose("verbose")
    logger.debug("debug")

    output = capsys.readouterr().out.splitlines()

    assert "info" in output
    assert "verbose" in output
    assert "debug" in output


def test_capture_stacktrace():
    """capture_stacktrace sets Log.stacktrace."""
    logger = Log()
    assert logger.skip_log is False

    try:
        1 / 0
    except ZeroDivisionError:
        logger.capture_stacktrace()

    assert len(logger.stacktraces) == 1
    assert logger.stacktraces[0][0] == "Main thread"
    assert isinstance(logger.stacktraces[0][1], Trace)
    assert logger.skip_log is False


@pytest.mark.parametrize("skip_logfile", [True, False])
def test_capture_stacktrace_for_briefcaseerror(skip_logfile):
    """skip_log is updated for BriefcaseError exceptions."""
    logger = Log()
    assert logger.skip_log is False

    try:
        raise BriefcaseError(error_code=542, skip_logfile=skip_logfile)
    except BriefcaseError:
        logger.capture_stacktrace()

    assert len(logger.stacktraces) == 1
    assert logger.stacktraces[0][0] == "Main thread"
    assert isinstance(logger.stacktraces[0][1], Trace)
    assert logger.skip_log is skip_logfile


def test_save_log_to_file_do_not_log(command):
    """Nothing is done to save log if no command or --log wasn't passed."""
    logger = Log()
    logger.save_log_to_file(command=None)

    logger.save_log = False
    logger.save_log_to_file(command=command)
    command.input.wait_bar.assert_not_called()

    # There were no stack traces captured
    assert len(logger.stacktraces) == 0


def test_save_log_to_file_no_exception(mock_now, command, tmp_path, monkeypatch):
    """Log file contains everything printed to log; env vars are sanitized; no
    stacktrace if one is not captured."""
    command.tools.os.environ = {
        "GITHUB_KEY": "super-secret-key",
        "ANDROID_HOME": "/androidsdk",
    }

    printer = Printer()
    command.tools.input = Console(printer=printer)
    logger = command.tools.logger = Log(printer=printer, verbosity=LogLevel.DEBUG)
    logger.save_log = True

    logger.debug("this is debug output")
    logger.info("this is info output")
    logger.warning("this is warning output")
    logger.error("this is error output")
    logger.print("this is print output")
    logger.print.to_log("this is log output")
    logger.print.to_log(f"{chr(7)}this is sanitized log output: \u001b[31mred")
    logger.print.to_console("this is console output")

    logger.info("this is [bold]info output with markup[/bold]")
    logger.info(
        "this is [bold]info output with markup and a prefix[/bold]", prefix="wibble"
    )
    logger.info("this is [bold]info output with escaped markup[/bold]", markup=True)
    logger.info(
        "this is [bold]info output with escaped markup and a prefix[/bold]",
        prefix="wibble",
        markup=True,
    )

    with command.tools.input.wait_bar("transient message...", transient=True):
        pass
    with command.tools.input.wait_bar("wait message..."):
        pass
    with contextlib.suppress(KeyboardInterrupt):
        with command.tools.input.wait_bar("abort message..."):
            raise KeyboardInterrupt

    project_root = tmp_path / "project_root"
    project_root.mkdir(exist_ok=True)
    monkeypatch.chdir(project_root)
    with open("pyproject.toml", "w", encoding="utf-8") as f:
        f.writelines(["[section]\n", "name = 'project'\n\n\n\n\n"])

    logger.save_log_to_file(command=command)

    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"

    assert log_filepath.exists()
    log_contents = log_filepath.read_text(encoding="utf-8")

    assert log_contents.startswith("Date/Time:       2022-06-25 16:12:29")
    assert "this is debug output" in log_contents
    assert "this is info output" in log_contents
    assert "this is [bold]info output with markup[/bold]" in log_contents
    assert "this is info output with escaped markup" in log_contents
    assert "this is warning output" in log_contents
    assert "this is error output" in log_contents
    assert "this is print output" in log_contents
    assert "this is log output" in log_contents
    assert "this is sanitized log output: red" in log_contents
    assert "this is console output" not in log_contents
    # Environment variables are in the output
    assert "ANDROID_HOME=/androidsdk" in log_contents
    assert "GITHUB_KEY=********************" in log_contents
    assert "GITHUB_KEY=super-secret-key" not in log_contents
    # Environment variables are sorted
    assert log_contents.index("ANDROID_HOME") < log_contents.index("GITHUB_KEY")
    # wait bar status messages are included
    assert "transient message... started" in log_contents
    assert "transient message... done" in log_contents
    assert "wait message... started" in log_contents
    assert "wait message... done" in log_contents
    assert "abort message... started" in log_contents
    assert "abort message... aborted" in log_contents
    assert "pyproject.toml:\n[section]\nname = 'project'\n\nBriefcase" in log_contents

    assert TRACEBACK_HEADER not in log_contents
    assert EXTRA_HEADER not in log_contents


def test_save_log_to_file_with_exception(mock_now, command, tmp_path):
    """Log file contains exception stacktrace when one is captured."""
    logger = Log()
    logger.save_log = True
    try:
        1 / 0
    except ZeroDivisionError:
        logger.capture_stacktrace()
    logger.save_log_to_file(command=command)

    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"

    assert log_filepath.exists()
    log_contents = log_filepath.read_text(encoding="utf-8")

    assert len(logger.stacktraces) == 1
    assert log_contents.startswith("Date/Time:       2022-06-25 16:12:29")
    assert TRACEBACK_HEADER in log_contents
    assert log_contents.splitlines()[-1].startswith("ZeroDivisionError")


def test_save_log_to_file_with_multiple_exceptions(mock_now, command, tmp_path):
    """Log file contains exception stacktrace when more than one is captured."""
    logger = Log()
    logger.save_log = True
    for i in range(1, 5):
        try:
            1 / 0
        except ZeroDivisionError:
            logger.capture_stacktrace(f"Thread {i}")

    logger.save_log_to_file(command=command)

    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"

    assert log_filepath.exists()
    log_contents = log_filepath.read_text(encoding="utf-8")

    assert len(logger.stacktraces) == 4
    assert log_contents.startswith("Date/Time:       2022-06-25 16:12:29")
    assert TRACEBACK_HEADER in log_contents
    for i in range(1, 5):
        assert f"\nThread {i} traceback:\n" in log_contents
    assert log_contents.splitlines()[-1].startswith("ZeroDivisionError")


def test_save_log_to_file_extra(mock_now, command, tmp_path):
    """Log file extras are called when the log is written."""
    logger = Log()
    logger.save_log = True

    def extra1():
        logger.debug("Log extra 1")

    def extra2():
        raise ValueError("Log extra 2")

    def extra3():
        logger.debug("Log extra 3")

    for extra in [extra1, extra2, extra3]:
        logger.add_log_file_extra(extra)
    logger.save_log_to_file(command=command)
    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"
    log_contents = log_filepath.read_text(encoding="utf-8")

    assert EXTRA_HEADER in log_contents
    assert "Log extra 1" in log_contents
    assert TRACEBACK_HEADER in log_contents
    assert "ValueError: Log extra 2" in log_contents
    assert "Log extra 3" in log_contents


def test_save_log_to_file_extra_interrupted(mock_now, command, tmp_path):
    """Log file extras can be interrupted by Ctrl-C."""
    logger = Log()
    logger.save_log = True

    def extra1():
        raise KeyboardInterrupt()

    extra2 = MagicMock()
    for extra in [extra1, extra2]:
        logger.add_log_file_extra(extra)
    with pytest.raises(KeyboardInterrupt):
        logger.save_log_to_file(command=command)
    extra2.assert_not_called()
    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"
    assert log_filepath.stat().st_size == 0


def test_save_log_to_file_missing_pyproject(mock_now, command, tmp_path, monkeypatch):
    """Log file contains pyproject read exception if it's missing."""
    logger = Log()
    logger.save_log = True

    # ensure in a directory without a pyproject.toml
    monkeypatch.chdir(tmp_path)
    Path(tmp_path / "pyproject.toml").unlink(missing_ok=True)

    logger.save_log_to_file(command=command)

    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"

    assert log_filepath.exists()
    log_contents = log_filepath.read_text(encoding="utf-8")

    assert "pyproject.toml:\n[Errno 2] No such file or directory" in log_contents


def test_save_log_to_file_fail_to_make_logs_dir(
    mock_now,
    command,
    capsys,
    monkeypatch,
    tmp_path,
):
    """User is informed when the ``logs`` directory cannot be created."""
    # Mock the command's base path such that it:
    mock_base_path = MagicMock(wraps=tmp_path)
    command.base_path = mock_base_path
    #  - returns a mocked filepath for the log file
    mock_base_path.__str__.return_value = "/asdf/log_filepath"
    #  - returns itself when the "logs" directory and log filename are appended
    #    to create the full filepath to the log file
    mock_base_path.__truediv__.return_value = mock_base_path
    #  - returns itself when ``Log`` requests the ``parent`` for the log filepath
    type(mock_base_path).parent = PropertyMock(return_value=mock_base_path)
    #  - raises for the call to ``mkdir`` to create the ``logs`` directory
    mock_base_path.mkdir.side_effect = OSError("directory creation denied")

    logger = Log()
    logger.save_log = True

    logger.print("a line of output")
    logger.save_log_to_file(command=command)

    assert capsys.readouterr().out == "\n".join(
        [
            "a line of output",
            "",
            "Failed to save log to /asdf/log_filepath: directory creation denied",
            "",
            "",
        ]
    )


def test_save_log_to_file_fail_to_write_file(
    mock_now,
    command,
    capsys,
    monkeypatch,
    tmp_path,
):
    """User is informed when the log file cannot be written."""
    # Mock opening a file that raises PermissionError on write
    mock_open = MagicMock(spec_set=open)
    monkeypatch.setattr("builtins.open", mock_open)
    mock_log_file = MagicMock(spec_set=TextIOBase)
    mock_open.return_value.__enter__.return_value = mock_log_file
    mock_log_file.write.side_effect = OSError("file write denied")

    logger = Log()
    logger.save_log = True

    logger.print("a line of output")
    logger.save_log_to_file(command=command)

    log_filepath = tmp_path / "logs/briefcase.2022_06_25-16_12_29.dev.log"
    assert capsys.readouterr().out == "\n".join(
        [
            "a line of output",
            "",
            f"Failed to save log to {log_filepath}: file write denied",
            "",
            "",
        ]
    )


def test_log_with_context(capsys):
    """Log file can be given a persistent context."""
    logger = Log(verbosity=LogLevel.DEBUG)
    logger.save_log = False

    logger.info("this is info output")
    with logger.context("Deep"):
        logger.info("this is deep context")
        logger.info("prefixed deep context", prefix="prefix")
        logger.info()
        logger.debug("this is deep debug")
        with logger.context("Really Deep"):
            logger.info("this is really deep context")
            logger.info("prefixed really deep context", prefix="prefix2")
            logger.info()
            logger.debug("this is really deep debug")
        logger.info("Pop back to deep")
    logger.info("Pop back to normal")

    assert capsys.readouterr().out == "\n".join(
        [
            "this is info output",
            "",
            "Entering Deep context...",
            "Deep| --------------------------------------------------------------------",
            "Deep| this is deep context",
            "Deep| ",
            "Deep| [prefix] prefixed deep context",
            "Deep| ",
            "Deep| this is deep debug",
            "Deep| ",
            "Deep| Entering Really Deep context...",
            "Really Deep| -------------------------------------------------------------",
            "Really Deep| this is really deep context",
            "Really Deep| ",
            "Really Deep| [prefix2] prefixed really deep context",
            "Really Deep| ",
            "Really Deep| this is really deep debug",
            "Really Deep| -------------------------------------------------------------",
            "Deep| Leaving Really Deep context.",
            "Deep| ",
            "Deep| Pop back to deep",
            "Deep| --------------------------------------------------------------------",
            "Leaving Deep context.",
            "",
            "Pop back to normal",
            "",
        ]
    )


def test_log_error_with_context(capsys):
    """If an exception is raised in a logging context, the context is cleared."""
    logger = Log(verbosity=LogLevel.DEBUG)
    logger.save_log = False

    logger.info("this is info output")
    try:
        with logger.context("Deep"):
            logger.info("this is deep context")
            raise ValueError()
    except ValueError:
        logger.info("this is cleanup")

    assert capsys.readouterr().out == "\n".join(
        [
            "this is info output",
            "",
            "Entering Deep context...",
            "Deep| --------------------------------------------------------------------",
            "Deep| this is deep context",
            "Deep| --------------------------------------------------------------------",
            "Leaving Deep context.",
            "",
            "this is cleanup",
            "",
        ]
    )


@pytest.mark.parametrize(
    "logging_level, handler_expected",
    [
        (LogLevel.DEEP_DEBUG, True),
        (LogLevel.DEBUG, False),
        (LogLevel.VERBOSE, False),
        (LogLevel.INFO, False),
    ],
)
def test_stdlib_logging_config(logging_level, handler_expected, logging_logger):
    """A logging handler is only added for DEEP_DEBUG mode."""
    logger = Log(verbosity=logging_level)

    logger.configure_stdlib_logging("test_pkg")

    assert handler_expected is any(
        isinstance(h, RichLoggingHandler) for h in logging_logger.handlers
    )


def test_stdlib_logging_only_one(logging_logger):
    """Only one logging handler is ever created for a package."""
    logger = Log(verbosity=LogLevel.DEEP_DEBUG)

    logger.configure_stdlib_logging("test_pkg")
    logger.configure_stdlib_logging("test_pkg")
    logger.configure_stdlib_logging("test_pkg")

    assert len(logging_logger.handlers) == 1


def test_stdlib_logging_handler_writes_to_debug(logging_logger):
    """The logging handler writes to the console through Log()."""
    logger = Log(verbosity=LogLevel.DEEP_DEBUG)
    logger.debug = MagicMock(wraps=logger.debug)

    logger.configure_stdlib_logging("test_pkg")

    logging_logger.debug("This is debug output")
    logging_logger.info("This is info output")

    assert logger.debug.mock_calls == [
        call("DEBUG test_pkg: This is debug output\n"),
        call("INFO test_pkg: This is info output\n"),
    ]



import pytest

from briefcase.exceptions import BriefcaseCommandError


def test_specific_app(build_command, first_app, second_app):
    """If a specific app is requested, build it."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = build_command.parse_options([])

    # Run the build command
    build_command(first_app, **options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App config has been finalized
        ("finalize-app-config", "first"),
        # App template is verified
        ("verify-app-template", "first"),
        # App tools are verified for app
        ("verify-app-tools", "first"),
        # Build the first app; no state
        ("build", "first", {"test_mode": False}),
    ]


def test_multiple_apps(build_command, first_app, second_app):
    """If there are multiple apps, build all of them."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = build_command.parse_options([])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # Build the first app; no state
        ("build", "first", {"test_mode": False}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Build the second apps; state from previous build.
        ("build", "second", {"build_state": "first", "test_mode": False}),
    ]


def test_non_existent(build_command, first_app_config, second_app):
    """Requesting a build of a non-existent app causes a create."""
    # Add two apps; use the "config only" version of the first app.
    build_command.apps = {
        "first": first_app_config,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = build_command.parse_options([])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First App doesn't exist, so it will be created, then built
        ("create", "first", {"test_mode": False}),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"create_state": "first", "test_mode": False}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Second app *does* exist, so it only be built
        (
            "build",
            "second",
            {"create_state": "first", "build_state": "first", "test_mode": False},
        ),
    ]


def test_unbuilt(build_command, first_app_unbuilt, second_app):
    """Requesting a build of an app that has been created, but not build, just causes a
    build."""
    # Add two apps; use the "unbuilt" version of the first app.
    build_command.apps = {
        "first": first_app_unbuilt,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = build_command.parse_options([])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        # First App exists, but hasn't been built; it will be built.
        ("build", "first", {"test_mode": False}),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        # Second app has been built before; it will be built again.
        ("build", "second", {"build_state": "first", "test_mode": False}),
    ]


def test_update_app(build_command, first_app, second_app):
    """If an update is requested, app is updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure a -a command line option
    options, _ = build_command.parse_options(["-u"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": False}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": False},
        ),
    ]


def test_update_app_requirements(build_command, first_app, second_app):
    """If a requirements update is requested, app is updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure update command line options
    options, _ = build_command.parse_options(["-r"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": False,
                "update_requirements": True,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": False}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": True,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": False},
        ),
    ]


def test_update_app_support(build_command, first_app, second_app):
    """If a support update is requested, support is updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure update command line options
    options, _ = build_command.parse_options(["--update-support"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": True,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": False}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": True,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": False},
        ),
    ]


def test_update_app_stub(build_command, first_app, second_app):
    """If a stub update is requested, the stub is updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure update command line options
    options, _ = build_command.parse_options(["--update-stub"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": True,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": False}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": True,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": False},
        ),
    ]


def test_update_app_resources(build_command, first_app, second_app):
    """If a resources update is requested, app is updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure update command line options
    options, _ = build_command.parse_options(["--update-resources"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": False,
                "update_requirements": False,
                "update_resources": True,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": False}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": False,
                "update_resources": True,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": False},
        ),
    ]


def test_update_non_existent(build_command, first_app_config, second_app):
    """Requesting an update of a non-existent app causes a create."""
    # Add two apps; use the "config only" version of the first app.
    build_command.apps = {
        "first": first_app_config,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = build_command.parse_options(["-u"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First App doesn't exist, so it will be created, then built
        ("create", "first", {"test_mode": False}),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"create_state": "first", "test_mode": False}),
        # Second app *does* exist, so it will be updated, then built
        (
            "update",
            "second",
            {
                "create_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {
                "create_state": "first",
                "build_state": "first",
                "update_state": "second",
                "test_mode": False,
            },
        ),
    ]


def test_update_unbuilt(build_command, first_app_unbuilt, second_app):
    """Requesting an update of an unbuilt app causes an update before build."""
    # Add two apps; use the "unbuilt" version of the first app.
    build_command.apps = {
        "first": first_app_unbuilt,
        "second": second_app,
    }

    # Configure no command line options
    options, _ = build_command.parse_options(["-u"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First App exists, but hasn't been built; it will be updated then built.
        (
            "update",
            "first",
            {
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": False}),
        # Second app has been built before; it will be built again.
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": False,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": False},
        ),
    ]


def test_build_test(build_command, first_app, second_app):
    """If the user builds a test app, app is updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--test"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": True}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": True},
        ),
    ]


def test_build_test_no_update(build_command, first_app, second_app):
    """If the user builds a test app without app updates, requirements and resources are
    still updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--test", "--no-update"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # No update of the first app
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"test_mode": True}),
        # No update of the second app
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"build_state": "first", "test_mode": True},
        ),
    ]


def test_build_test_update_dependencies(build_command, first_app, second_app):
    """If the user builds a test app with app dependency updates, app code and resources
    are updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--test", "-r"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": True,
                "update_requirements": True,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": True}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": True,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": True},
        ),
    ]


def test_build_test_update_resources(build_command, first_app, second_app):
    """If the user builds a test app with app resource updates, app code and resources
    are updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--test", "--update-resources"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": True,
                "update_requirements": False,
                "update_resources": True,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": True}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": False,
                "update_resources": True,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": True},
        ),
    ]


def test_build_test_update_support(build_command, first_app, second_app):
    """If the user builds a test app with a support update, app code and support are
    updated before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--test", "--update-support"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": True,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": True}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": True,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": True},
        ),
    ]


def test_build_test_update_stub(build_command, first_app, second_app):
    """If the user builds a test app with stub update, app code and stub are updated
    before build."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--test", "--update-stub"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # Update then build the first app
        (
            "update",
            "first",
            {
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": True,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"update_state": "first", "test_mode": True}),
        # Update then build the second app
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": True,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": True},
        ),
    ]


def test_build_invalid_update(build_command, first_app, second_app):
    """If the user requests a build with update and no-update, an error is raised."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["-u", "--no-update"])

    # Run the build command
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Cannot specify both --update and --no-update",
    ):
        build_command(**options)


def test_build_invalid_update_requirements(build_command, first_app, second_app):
    """If the user requests a build with update-requirements and no-update, an error is
    raised."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["-r", "--no-update"])

    # Run the build command
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Cannot specify both --update-requirements and --no-update",
    ):
        build_command(**options)


def test_build_invalid_update_resources(build_command, first_app, second_app):
    """If the user requests a build with update-resources and no-update, an error is
    raised."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--update-resources", "--no-update"])

    # Run the build command
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Cannot specify both --update-resources and --no-update",
    ):
        build_command(**options)


def test_build_invalid_update_support(build_command, first_app, second_app):
    """If the user requests a build with update-support and no-update, an error is
    raised."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--update-support", "--no-update"])

    # Run the build command
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Cannot specify both --update-support and --no-update",
    ):
        build_command(**options)


def test_build_invalid_update_stub(build_command, first_app, second_app):
    """If the user requests a build with update-stub and no-update, an error is
    raised."""
    # Add two apps
    build_command.apps = {
        "first": first_app,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["--update-stub", "--no-update"])

    # Run the build command
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Cannot specify both --update-stub and --no-update",
    ):
        build_command(**options)


def test_test_app_non_existent(build_command, first_app_config, second_app):
    """Requesting a test build of a non-existent app causes a create."""
    # Add two apps; use the "config only" version of the first app.
    build_command.apps = {
        "first": first_app_config,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["-u", "--test"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First App doesn't exist, so it will be created, then built
        ("create", "first", {"test_mode": True}),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        ("build", "first", {"create_state": "first", "test_mode": True}),
        # Second app *does* exist, so it will be updated, then built
        (
            "update",
            "second",
            {
                "create_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {
                "create_state": "first",
                "build_state": "first",
                "update_state": "second",
                "test_mode": True,
            },
        ),
    ]


def test_test_app_unbuilt(build_command, first_app_unbuilt, second_app):
    """Requesting a test build with update of an unbuilt app causes an update before
    build."""
    # Add two apps; use the "unbuilt" version of the first app.
    build_command.apps = {
        "first": first_app_unbuilt,
        "second": second_app,
    }

    # Configure command line options
    options, _ = build_command.parse_options(["-u", "--test"])

    # Run the build command
    build_command(**options)

    # The right sequence of things will be done
    assert build_command.actions == [
        # Host OS is verified
        ("verify-host",),
        # Tools are verified
        ("verify-tools",),
        # App configs have been finalized
        ("finalize-app-config", "first"),
        ("finalize-app-config", "second"),
        # First App exists, but hasn't been built; it will be updated then built.
        (
            "update",
            "first",
            {
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for first app
        ("verify-app-template", "first"),
        # App tools are verified for first app
        ("verify-app-tools", "first"),
        (
            "build",
            "first",
            {"update_state": "first", "test_mode": True},
        ),
        # Second app has been built before; it will be built again.
        (
            "update",
            "second",
            {
                "update_state": "first",
                "build_state": "first",
                "test_mode": True,
                "update_requirements": False,
                "update_resources": False,
                "update_support": False,
                "update_stub": False,
            },
        ),
        # App template is verified for second app
        ("verify-app-template", "second"),
        # App tools are verified for second app
        ("verify-app-tools", "second"),
        (
            "build",
            "second",
            {"update_state": "second", "build_state": "first", "test_mode": True},
        ),
    ]



import os
import shutil
from unittest import mock

import pytest

import briefcase
from briefcase.exceptions import MissingAppSources

from ...utils import create_file


def assert_dist_info(app_path):
    dist_info_path = app_path / "my_app-1.2.3.dist-info"

    # Confirm the metadata files exist.
    assert (dist_info_path / "INSTALLER").exists()
    assert (dist_info_path / "METADATA").exists()

    with (dist_info_path / "INSTALLER").open(encoding="utf-8") as f:
        assert f.read() == "briefcase\n"

    with (dist_info_path / "METADATA").open(encoding="utf-8") as f:
        assert (
            f.read()
            == f"""Metadata-Version: 2.1
Briefcase-Version: {briefcase.__version__}
Name: my-app
Formal-Name: My App
App-ID: com.example.my-app
Version: 1.2.3
Home-page: https://example.com
Download-URL: https://example.com
Author: First Last
Author-email: first@example.com
Summary: This is a simple app
"""
        )


def test_no_code(
    create_command,
    myapp,
    app_path,
    app_requirements_path_index,
):
    """If an app has no code (?!), install_app_code is mostly a no-op; but distinfo is
    created."""
    # Mock shutil so we can track usage.
    create_command.tools.shutil = mock.MagicMock(spec_set=shutil)
    create_command.tools.os = mock.MagicMock(spec_set=os)

    myapp.sources = None

    create_command.install_app_code(myapp, test_mode=False)

    # No request was made to install requirements
    create_command.tools.shutil.rmtree.assert_called_once_with(app_path)
    create_command.tools.os.mkdir.assert_called_once_with(app_path)
    create_command.tools.shutil.copytree.assert_not_called()
    create_command.tools.shutil.copy.assert_not_called()

    # Metadata has been created
    assert_dist_info(app_path)


def test_empty_code(
    create_command,
    myapp,
    app_path,
    app_requirements_path_index,
):
    """If an app has an empty sources list (?!), install_app_code is mostly a no-op; but
    distinfo is created."""
    # Mock shutil so we can track usage.
    create_command.tools.shutil = mock.MagicMock(spec_set=shutil)
    create_command.tools.os = mock.MagicMock(spec_set=os)

    myapp.sources = []

    create_command.install_app_code(myapp, test_mode=False)

    # No request was made to install requirements
    create_command.tools.shutil.rmtree.assert_called_once_with(app_path)
    create_command.tools.os.mkdir.assert_called_once_with(app_path)
    create_command.tools.shutil.copytree.assert_not_called()
    create_command.tools.shutil.copy.assert_not_called()

    # Metadata has been created
    assert_dist_info(app_path)


def test_source_missing(
    create_command,
    myapp,
    app_path,
    app_requirements_path_index,
):
    """If an app defines sources that are missing, an error is raised."""
    # Set the app definition to point at sources that don't exist
    myapp.sources = ["missing"]

    with pytest.raises(MissingAppSources):
        create_command.install_app_code(myapp, test_mode=False)

    # Distinfo won't be created.
    dist_info_path = app_path / "myapp-1.2.3.dist-info"
    assert not dist_info_path.exists()


def test_source_dir(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """If an app defines directories of sources, the whole directory is copied."""
    # Create the mock sources
    # src /
    #   first /
    #     demo.py
    #   second /
    #     shallow.py
    #     submodule /
    #       deeper.py
    create_file(
        tmp_path / "base_path/src/first/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/shallow.py",
        "print('hello shallow second')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/submodule/deeper.py",
        "print('hello deep second')\n",
    )

    # Set the app definition, and install sources
    myapp.sources = ["src/first", "src/second"]

    create_command.install_app_code(myapp, test_mode=False)

    # All the sources exist.
    assert (app_path / "first").exists()
    assert (app_path / "first/demo.py").exists()

    assert (app_path / "second").exists()
    assert (app_path / "second/shallow.py").exists()

    assert (app_path / "second/submodule").exists()
    assert (app_path / "second/submodule/deeper.py").exists()

    # Metadata has been created
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources == ["src/first", "src/second"]
    assert myapp.test_sources is None


def test_source_file(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """If an app defines single file sources, the files are copied."""
    # Create the mock sources
    # src /
    #   demo.py
    # other.py
    create_file(
        tmp_path / "base_path/src/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/other.py",
        "print('hello second')\n",
    )

    # Set the app definition, and install sources
    myapp.sources = ["src/demo.py", "other.py"]

    create_command.install_app_code(myapp, test_mode=False)

    # All the sources exist.
    assert (app_path / "demo.py").exists()
    assert (app_path / "other.py").exists()

    # Metadata has been created
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources == ["src/demo.py", "other.py"]
    assert myapp.test_sources is None


def test_no_existing_app_folder(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """If there's no pre-existing app folder, one is created."""
    # Create the mock sources
    # src /
    #   first /
    #     demo.py
    #   second /
    #     shallow.py
    #     submodule /
    #       deeper.py
    create_file(
        tmp_path / "base_path/src/first/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/shallow.py",
        "print('hello shallow second')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/submodule/deeper.py",
        "print('hello deep second')\n",
    )

    # Remove the app folder created by the test fixture.
    shutil.rmtree(app_path)

    # Set the app definition, and install sources
    myapp.sources = ["src/first/demo.py", "src/second"]

    create_command.install_app_code(myapp, test_mode=False)

    # All the new sources exist, and contain the new content.
    assert (app_path / "demo.py").exists()
    with (app_path / "demo.py").open(encoding="utf-8") as f:
        assert f.read() == "print('hello first')\n"

    assert (app_path / "second").exists()
    assert (app_path / "second/shallow.py").exists()
    with (app_path / "second/shallow.py").open(encoding="utf-8") as f:
        assert f.read() == "print('hello shallow second')\n"

    assert (app_path / "second/submodule").exists()
    assert (app_path / "second/submodule/deeper.py").exists()
    with (app_path / "second/submodule/deeper.py").open(encoding="utf-8") as f:
        assert f.read() == "print('hello deep second')\n"

    # The stale/broken modules have been removed.
    assert not (app_path / "stale.py").exists()
    assert not (app_path / "second/stale.py").exists()
    assert not (app_path / "second/broken").exists()

    # Metadata has been updated.
    assert not (app_path / "my_app-1.2.2.dist-info").exists()
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources == ["src/first/demo.py", "src/second"]
    assert myapp.test_sources is None


def test_replace_sources(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """Stale sources and dist-info are removed on installation."""
    # Create the mock sources
    # src /
    #   first /
    #     demo.py
    #   second /
    #     shallow.py
    #     submodule /
    #       deeper.py
    create_file(
        tmp_path / "base_path/src/first/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/shallow.py",
        "print('hello shallow second')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/submodule/deeper.py",
        "print('hello deep second')\n",
    )

    # Also create some existing sources:
    # path / to / app /
    #   demo.py
    #   stale.py
    #   second /
    #     shallow.py
    #     stale.py
    #     submodule /
    #       deeper.py
    #     broken /
    #       other.py
    #   my_app-1.2.2.dist-info /
    create_file(
        app_path / "src/demo.py",
        "print('old hello first')\n",
    )
    create_file(
        app_path / "src/stale.py",
        "print('stale hello first')\n",
    )
    create_file(
        app_path / "src/second/shallow.py",
        "print('old hello shallow second')\n",
    )
    create_file(
        app_path / "src/second/stale.py",
        "print('hello second stale')\n",
    )
    create_file(
        app_path / "src/second/submodule/deeper.py",
        "print('hello deep second')\n",
    )
    create_file(
        app_path / "src/second/broken/other.py",
        "print('hello second deep broken')\n",
    )

    old_dist_info_dir = app_path / "my_app-1.2.2.dist-info"
    old_dist_info_dir.mkdir()

    # Set the app definition, and install sources
    myapp.sources = ["src/first/demo.py", "src/second"]

    create_command.install_app_code(myapp, test_mode=False)

    # All the new sources exist, and contain the new content.
    assert (app_path / "demo.py").exists()
    with (app_path / "demo.py").open(encoding="utf-8") as f:
        assert f.read() == "print('hello first')\n"

    assert (app_path / "second").exists()
    assert (app_path / "second/shallow.py").exists()
    with (app_path / "second/shallow.py").open(encoding="utf-8") as f:
        assert f.read() == "print('hello shallow second')\n"

    assert (app_path / "second/submodule").exists()
    assert (app_path / "second/submodule/deeper.py").exists()
    with (app_path / "second/submodule/deeper.py").open(encoding="utf-8") as f:
        assert f.read() == "print('hello deep second')\n"

    # The stale/broken modules have been removed.
    assert not (app_path / "stale.py").exists()
    assert not (app_path / "second/stale.py").exists()
    assert not (app_path / "second/broken").exists()

    # Metadata has been updated.
    assert not (app_path / "my_app-1.2.2.dist-info").exists()
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources == ["src/first/demo.py", "src/second"]
    assert myapp.test_sources is None


def test_non_latin_metadata(
    create_command,
    myapp,
    app_path,
    app_requirements_path_index,
):
    """If the app metadata contains non-Latin-1 characters, the METADATA file is written
    correctly (Briefcase#767)"""
    myapp.formal_name = "My büggy app"
    myapp.author = "José Weiß-Müller"
    myapp.author_email = "钱华林@中科院.中国"
    myapp.url = "https://xn--7qvx15a.cn"
    myapp.description = "A Møøse once bit my sister..."

    # Mock shutil so we can track usage.
    create_command.tools.shutil = mock.MagicMock(spec_set=shutil)
    create_command.tools.os = mock.MagicMock(spec_set=os)

    myapp.sources = []

    create_command.install_app_code(myapp, test_mode=False)

    # No request was made to install requirements
    create_command.tools.shutil.rmtree.assert_called_once_with(app_path)
    create_command.tools.os.mkdir.assert_called_once_with(app_path)
    create_command.tools.shutil.copytree.assert_not_called()
    create_command.tools.shutil.copy.assert_not_called()

    # The dist-info file was created, and is readable.
    dist_info_path = app_path / "my_app-1.2.3.dist-info"

    # Confirm the metadata files exist.
    assert (dist_info_path / "INSTALLER").exists()
    assert (dist_info_path / "METADATA").exists()

    with (dist_info_path / "INSTALLER").open(encoding="utf-8") as f:
        assert f.read() == "briefcase\n"

    with (dist_info_path / "METADATA").open(encoding="utf-8") as f:
        assert (
            f.read()
            == f"""Metadata-Version: 2.1
Briefcase-Version: {briefcase.__version__}
Name: my-app
Formal-Name: My büggy app
App-ID: com.example.my-app
Version: 1.2.3
Home-page: https://xn--7qvx15a.cn
Download-URL: https://xn--7qvx15a.cn
Author: José Weiß-Müller
Author-email: 钱华林@中科院.中国
Summary: A Møøse once bit my sister...
"""
        )


def test_test_sources(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """If an app defines test code, but we're not in test mode, it isn't copied."""
    # Create the mock sources
    # src /
    #   first /
    #     demo.py
    #   second /
    #     shallow.py
    # tests /
    #   first.py
    #   deep /
    #     test_case.py
    # othertests/
    #   tests_more.py
    #   special /
    #     test_weird.py
    create_file(
        tmp_path / "base_path/src/first/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/shallow.py",
        "print('hello shallow second')\n",
    )
    create_file(
        tmp_path / "base_path/tests/first.py",
        "print('hello first test suite')\n",
    )
    create_file(
        tmp_path / "base_path/tests/deep/test_case.py",
        "print('hello test case')\n",
    )
    create_file(
        tmp_path / "base_path/othertests/test_more.py",
        "print('hello more tests')\n",
    )
    create_file(
        tmp_path / "base_path/othertests/special/test_weird.py",
        "print('hello weird tests')\n",
    )

    # Set the app definition, and install sources
    myapp.sources = ["src/first", "src/second"]
    myapp.test_sources = ["tests", "othertests"]

    create_command.install_app_code(myapp, test_mode=False)

    # App sources exist.
    assert (app_path / "first").exists()
    assert (app_path / "first/demo.py").exists()

    assert (app_path / "second").exists()
    assert (app_path / "second/shallow.py").exists()

    # Test sources do not exist
    assert not (app_path / "tests").exists()
    assert not (app_path / "othertests").exists()

    # Metadata has been created
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources == ["src/first", "src/second"]
    assert myapp.test_sources == ["tests", "othertests"]


def test_test_sources_test_mode(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """If an app defines test code, and we're in test mode, test sources are copied."""
    # Create the mock sources
    # src /
    #   first /
    #     demo.py
    #   second /
    #     shallow.py
    # tests /
    #   first.py
    #   deep /
    #     test_case.py
    # othertests/
    #   tests_more.py
    #   special /
    #     test_weird.py
    create_file(
        tmp_path / "base_path/src/first/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/shallow.py",
        "print('hello shallow second')\n",
    )
    create_file(
        tmp_path / "base_path/tests/first.py",
        "print('hello first test suite')\n",
    )
    create_file(
        tmp_path / "base_path/tests/deep/test_case.py",
        "print('hello test case')\n",
    )
    create_file(
        tmp_path / "base_path/othertests/test_more.py",
        "print('hello more tests')\n",
    )
    create_file(
        tmp_path / "base_path/othertests/special/test_weird.py",
        "print('hello weird tests')\n",
    )

    # Set the app definition, and install sources
    myapp.sources = ["src/first", "src/second"]
    myapp.test_sources = ["tests", "othertests"]

    create_command.install_app_code(myapp, test_mode=True)

    # App sources exist.
    assert (app_path / "first").exists()
    assert (app_path / "first/demo.py").exists()

    assert (app_path / "second").exists()
    assert (app_path / "second/shallow.py").exists()

    # Test sources exist
    assert (app_path / "tests/first.py").exists()
    assert (app_path / "tests/deep/test_case.py").exists()

    assert (app_path / "othertests/test_more.py").exists()
    assert (app_path / "othertests/special/test_weird.py").exists()

    # Metadata has been created
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources == ["src/first", "src/second"]
    assert myapp.test_sources == ["tests", "othertests"]


def test_only_test_sources_test_mode(
    create_command,
    myapp,
    tmp_path,
    app_path,
    app_requirements_path_index,
):
    """If an app only defines test code, and we're in test mode, test sources are
    copied."""
    # Create the mock sources
    # tests /
    #   first.py
    #   deep /
    #     test_case.py
    # othertests/
    #   tests_more.py
    #   special /
    #     test_weird.py
    create_file(
        tmp_path / "base_path/src/first/demo.py",
        "print('hello first')\n",
    )
    create_file(
        tmp_path / "base_path/src/second/shallow.py",
        "print('hello shallow second')\n",
    )
    create_file(
        tmp_path / "base_path/tests/first.py",
        "print('hello first test suite')\n",
    )
    create_file(
        tmp_path / "base_path/tests/deep/test_case.py",
        "print('hello test case')\n",
    )
    create_file(
        tmp_path / "base_path/othertests/test_more.py",
        "print('hello more tests')\n",
    )
    create_file(
        tmp_path / "base_path/othertests/special/test_weird.py",
        "print('hello weird tests')\n",
    )

    # Set the app definition, and install sources
    myapp.sources = None
    myapp.test_sources = ["tests", "othertests"]

    create_command.install_app_code(myapp, test_mode=True)

    # App sources do not exist.
    assert not (app_path / "first").exists()
    assert not (app_path / "second").exists()

    # Test sources exist
    assert (app_path / "tests/first.py").exists()
    assert (app_path / "tests/deep/test_case.py").exists()

    assert (app_path / "othertests/test_more.py").exists()
    assert (app_path / "othertests/special/test_weird.py").exists()

    # Metadata has been created
    assert_dist_info(app_path)

    # Original app definitions haven't changed
    assert myapp.sources is None
    assert myapp.test_sources == ["tests", "othertests"]



import pytest

from briefcase.exceptions import BriefcaseCommandError
from briefcase.platforms.macOS import SigningIdentity


@pytest.mark.parametrize(
    "identity_id, identity_name, team_id",
    [
        ("CAFEBEEF", "Developer ID Application: Jane Developer (DEADBEEF)", "DEADBEEF"),
        (
            "CAFEBEEF",
            "Developer ID Application: Edwin (Buzz) Aldrin (DEADBEEF)",
            "DEADBEEF",
        ),
    ],
)
def test_identity(identity_id, identity_name, team_id):
    """A signing identity can be created."""
    identity = SigningIdentity(id=identity_id, name=identity_name)
    assert identity.id == identity_id
    assert identity.name == identity_name
    assert identity.team_id == team_id
    assert not identity.is_adhoc
    assert repr(identity) == f"<SigningIdentity id={identity_id}>"


@pytest.mark.parametrize(
    "identity_name",
    [
        "Developer ID Application: Jane Developer",
        "DEADBEEF",
    ],
)
def test_bad_identity(identity_name):
    """Creating a bad identity raises an error."""
    with pytest.raises(
        BriefcaseCommandError,
        match=r"Couldn't extract Team ID from signing identity",
    ):
        SigningIdentity(id="CAFEBEEF", name=identity_name)


def test_adhoc_identity():
    """An ad-hoc identity can be created."""
    adhoc = SigningIdentity()
    assert adhoc.id == "-"
    assert (
        adhoc.name
        == "Ad-hoc identity. The resulting package will run but cannot be re-distributed."
    )
    assert adhoc.is_adhoc
    assert repr(adhoc) == "<AdhocSigningIdentity>"



from __future__ import annotations

import locale
import os
import platform
import shutil
import sys
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Mapping
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, DefaultDict, TypeVar

import requests
from cookiecutter.main import cookiecutter

from briefcase.config import AppConfig
from briefcase.console import Console, Log
from briefcase.exceptions import (
    MissingToolError,
    NonManagedToolError,
    UnsupportedHostError,
)

if TYPE_CHECKING:
    # Tools are imported only for type checking
    # to avoid circular import errors.
    import git as git_

    from briefcase.integrations.android_sdk import AndroidSDK
    from briefcase.integrations.docker import Docker, DockerAppContext
    from briefcase.integrations.file import File
    from briefcase.integrations.flatpak import Flatpak
    from briefcase.integrations.java import JDK
    from briefcase.integrations.linuxdeploy import LinuxDeploy
    from briefcase.integrations.rcedit import RCEdit
    from briefcase.integrations.subprocess import Subprocess
    from briefcase.integrations.visualstudio import VisualStudio
    from briefcase.integrations.windows_sdk import WindowsSDK
    from briefcase.integrations.wix import WiX
    from briefcase.integrations.xcode import Xcode, XcodeCliTools

ToolT = TypeVar("ToolT", bound="Tool")
ManagedToolT = TypeVar("ManagedToolT", bound="ManagedTool")

# Registry of all defined Tools
tool_registry: dict[str, type[Tool | ManagedTool]] = dict()

DEFAULT_SYSTEM_ENCODING = "UTF-8"


class Tool(ABC):
    """Tool Base."""

    name: str
    full_name: str
    supported_host_os: set[str] = {"Darwin", "Linux", "Windows"}

    def __init__(self, tools: ToolCache, **kwargs):
        self.tools = tools

    def __init_subclass__(cls, **kwargs):
        """Register each tool when it is defined."""
        if cls.name != "managed_tool_base":
            tool_registry[cls.name] = cls

    @classmethod
    def verify(
        cls: type[ToolT],
        tools: ToolCache,
        app: AppConfig | None = None,
        **kwargs,
    ) -> ToolT:
        """Confirm the tool is available and usable on the host platform."""
        cls.verify_host(tools=tools)
        tool = cls.verify_install(tools=tools, app=app, **kwargs)
        return tool

    @classmethod
    def verify_host(cls, tools: ToolCache):
        """Confirm the tool is supported on the platform."""
        if tools.host_os not in cls.supported_host_os:
            raise UnsupportedHostError(
                f"{cls.name} is not supported on {tools.host_os}"
            )

    @classmethod
    @abstractmethod
    def verify_install(cls: type[ToolT], tools: ToolCache, **kwargs) -> ToolT:
        """Confirm the tool is installed and available."""

    @property
    def managed_install(self) -> bool:
        """Is Briefcase managing the installation of this tool?"""
        return False


class ManagedTool(Tool):
    """Tool that can be managed by Briefcase."""

    name = "managed_tool_base"

    @classmethod
    def verify(
        cls: type[ManagedToolT],
        tools: ToolCache,
        app: AppConfig | None = None,
        install: bool = True,
        **kwargs,
    ) -> ManagedToolT:
        """Confirm the managed tool is installed and available."""
        return super().verify(tools=tools, app=app, install=install, **kwargs)

    @abstractmethod
    def exists(self) -> bool:
        """Is the tool currently installed?"""

    @abstractmethod
    def install(self):
        """Install the tool as managed by Briefcase."""

    @abstractmethod
    def uninstall(self):
        """Uninstall a Briefcase managed tool."""

    @property
    def managed_install(self) -> bool:
        """Is Briefcase managing the installation of this tool?"""
        return True

    def upgrade(self):
        """Upgrade a managed tool."""
        if self.managed_install:
            if not self.exists():
                raise MissingToolError(self.full_name)
            self.uninstall()
            self.install()
        else:
            raise NonManagedToolError(self.full_name)


class ToolCache(Mapping):
    # Useful fixed filesystem locations
    ETC_OS_RELEASE: Path = Path("/etc/os-release")

    # Briefcase tools
    android_sdk: AndroidSDK
    app_context: Subprocess | DockerAppContext
    docker: Docker
    file: File
    flatpak: Flatpak
    git: git_
    java: JDK
    linuxdeploy: LinuxDeploy
    rcedit: RCEdit
    subprocess: Subprocess
    visualstudio: VisualStudio
    windows_sdk: WindowsSDK
    wix: WiX
    xcode: Xcode
    xcode_cli: XcodeCliTools

    # Python stdlib tools
    platform = platform
    os = os
    shutil = shutil
    sys = sys

    # Third party tools
    cookiecutter = staticmethod(cookiecutter)
    requests = requests

    def __init__(
        self,
        logger: Log,
        console: Console,
        base_path: Path,
        home_path: Path | None = None,
    ):
        """Cache for managing tool access and verification.

        Non-app-specific tools are available via attribute access:
            e.g.: tools.subprocess
        App-specific tools are available via dictionary access:
            e.g.: tools[app].app_context

        :param logger: Logger for console and logfile.
        :param console: Facilitates console interaction and input solicitation.
        :param base_path: Base directory for tools (e.g. ~/.cache/briefcase/tools).
        :param home_path: Home directory for current user.
        """
        self.logger = logger
        self.input = console
        self.base_path = Path(base_path)
        self.home_path = Path(
            os.path.expanduser(home_path if home_path else Path.home())
        )

        self.host_arch = self.platform.machine()
        self.host_os = self.platform.system()
        # Python is 32bit if its pointers can only address with 32 bits or fewer
        self.is_32bit_python = self.sys.maxsize <= 2**32

        self.app_tools: DefaultDict[AppConfig, ToolCache] = defaultdict(
            lambda: ToolCache(
                logger=self.logger,
                console=self.input,
                base_path=self.base_path,
                home_path=self.home_path,
            )
        )

    @cached_property
    def system_encoding(self) -> str:
        """The character encoding for the system's locale.

        This locale API tries to determine the system's default encoding and generally
        works on typically configured systems; although, there are potential pitfalls
        in certain situations...so, this is best-effort.

        This API is used over getpreferredencoding() to avoid respecting Python's UTF-8
        mode; the system may not be using UTF-8 even if Python is configured to use it.

        :returns: a character encoding (upper-cased), e.g. UTF-8. Defaults to UTF-8.
        """
        if sys.version_info < (3, 11):  # pragma: no-cover-if-gte-py311
            encoding = locale.getdefaultlocale()[1]  # deprecated in Python 3.11
        else:  # pragma: no-cover-if-lt-py311
            encoding = locale.getencoding()

        if not encoding:
            encoding = DEFAULT_SYSTEM_ENCODING

        return encoding.upper()

    def __getitem__(self, app: AppConfig) -> ToolCache:
        return self.app_tools[app]

    def __iter__(self):
        return iter(self.app_tools)

    def __len__(self) -> int:
        return len(self.app_tools)

    def __bool__(self):
        # always True instead of __len__() != 0
        return True



import pytest

from briefcase.commands.new import parse_project_overrides
from briefcase.exceptions import BriefcaseCommandError


@pytest.mark.parametrize(
    "cmdline, overrides",
    [
        ([], {}),
        (["license=MIT"], {"license": "MIT"}),
        (["bootstrap=Toga Automation"], {"bootstrap": "Toga Automation"}),
        (["one=1", "two=2"], {"one": "1", "two": "2"}),
        (["one\n=\n1", "two=\n\n2"], {"one": "1", "two": "2"}),
        (["one==1", "two===2"], {"one": "=1", "two": "==2"}),
        (["   one  = 1  ", " two  =  2 "], {"one": "1", "two": "2"}),
    ],
)
def test_project_overrides(cmdline, overrides):
    """Valid project configuration overrides are parsed correctly."""
    assert parse_project_overrides(cmdline) == overrides


def test_project_overrides_invalid():
    """Invalid project configuration overrides are rejected."""
    with pytest.raises(
        BriefcaseCommandError,
        match="Unable to parse project configuration override ' license '",
    ):
        parse_project_overrides([" license "])


@pytest.mark.parametrize(
    "cmdline",
    [
        "=",
        " = ",
        "key_for_invalid_value=",
        " key_for_invalid_value =   ",
        "=value_for_invalid_key",
        "  = value_for_invalid_key   ",
    ],
)
def test_project_overrides_empty(cmdline):
    """Invalid project configuration overrides are rejected."""
    with pytest.raises(
        BriefcaseCommandError,
        match=f"Invalid Project configuration override '{cmdline}'",
    ):
        parse_project_overrides([cmdline])


def test_project_overrides_duplicate():
    """Duplicate project configuration overrides are rejected."""
    with pytest.raises(
        BriefcaseCommandError,
        match="Project configuration override 'license' specified multiple times",
    ):
        parse_project_overrides(["license=MIT", "license=BSD"])
